; **
; Intellivision for GameBoy Color v1.1
; Copyright(C) 2000,03 By Carl Mueller, Jr.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

; ---------------------------------------------------------------------------
;                    .::** Collision routines **::.
; ---------------------------------------------------------------------------

; **
; STIC_CollInit:   Setup every COLLMO record and field, including
;                  the CollAddr and BitParam fields.
;
; Parameters:      None.
;
; Modifies:        af, bc, hl, de
;
; Notes:           Must be run after the InitMO routine.
;

STIC_CollInit:

         ; first, copy the necessary routines into Low RAM

         ld             hl, STIC_CollChkMO_ROM
         ld             de, STIC_CollChkMO
         ld             bc, STIC_CollChkMO_ROM_End - STIC_CollChkMO_ROM
         call           MemCopy

         ld             hl, PixelPercise_ROM
         ld             de, PixelPercise
         ld             bc, PixelPercise_ROM_End - PixelPercise_ROM
         call           MemCopy

         ld             hl, BkgPixelPercise_ROM
         ld             de, BkgPixelPercise
         ld             bc, BkgPixelPercise_ROM_End - BkgPixelPercise_ROM
         call           MemCopy

         ; now initialize the COLLMO structure

         ld             hl, COLLMO
         ld             bc, 8 << 8 + $18 * 2
         ld             de, 19

         ld             a, $C6          ; a = MO # 0's BitParam
.Loop
         ld             [hl], a         ; store out CollFlag + MoSeven

         add            hl, de          ; pos COLLMO @ CollAddr

         ld             [hl], c         ; store out CollAddr ($30-$3F)
         inc            l
         ld             [hl+], a        ; store out BitParam ($C6-$FE)

         inc            c
         inc            c               ; advance to next CollAddr

         add            b               ; advance to next BitParam
         jr             nc, .Loop       ; no, so keep looping

         ld             [hl], d         ; clear "end-of-records" byte

         ld             [COLLMO + COLLMO_RECSIZE * 7], a  ; reset MoSeven

; **
; STIC_CollPrep:   Prepare COLLMO for each MO, taking into account
;                  its current state by examining the MO record
;                  structure.
;
; Parameters:      None.
;
; Modifies:        af, bc, hl, de
;
; Notes:           Must be called before STIC_Analyze so that
;                  "ReCopy" is still intact, and should be called
;                  before GRAM_CheckFlags, which may set ReCopy
;                  "unnecessarily" (doesn't have an affect on
;                  collision checking).
;

         ; CM = CollChkMo
         ; CB = CollChkBkg
         ; CS = CollChkScrn

         ;   Ofs   Name                Source       Used By      Dependencies
         ;
         ;   0     CollFlag + MoSeven  MO (calc'd)  CM, CB, CS
         ;   1     X-Position          MO           CM, CB, CS
         ; * 2     Width               MoAttrTab    CM, CB       XS
         ;   3     Y-Position          MO           CM, CB, CS
         ; * 4     Height              MoAttrTab    CM, CB, CS   YR, YS, YF
         ;   5     SrcOpCode           MoAttrTab    CM           YM
         ;   6     DstOpCode           MoAttrTab    CM           YM, YR
         ; * 7     X-Size Table        MoAttrTab    CM, CB       XS
         ;   8     X-Mirror Table      MoAttrTab    CM, CB       XM
         ; * 9     EntryPoint          MoAttrTab    CB           YS, YF
         ; * 10    LoopPoint           MoAttrTab    CB           YS, YF
         ;   11    OpCode #1           MoAttrTab    CB           YM, YS, YF
         ; * 12    OpCode #2           MoAttrTab    CB           YS, YF
         ;   13    OpCode #3           MoAttrTab    CB           YM, YS, YF
         ; * 14    AttrB1 (YS & YF)    MO           CM, CB, CS   YS, YF
         ;   15    AttrB2 (YR & YM)    MO           CM, CB, CS   YM, YR
         ; * 16    RepLine             RepLineTab   CM, CB       YS, YF
         ; * 17-18 DataAddr            MO           CM, CB, CS
         ;   19    CollAddr            CollInit     CM, CB, CS
         ;   20    BitParam            CollInit     CM

         ; * = only needs to be recalculated if ReCopy of MO attribute
         ;     byte #1 is set.

STIC_CollPrep:

         ld             de, MO
         ld             hl, COLLMO + 1  ; pos COLLMO @ X-Position
.Loop

         ; 1st, get "collisions-enabled" bit from MO attribute byte #0

         ld             a, [de]         ; a <- MO attribute byte #0
         inc            e

         and            $0D             ; isolate YM, XM and C
         ld             c, a            ; save them in c

         ; 2nd, get X-Position & allow collisions only if it is in 1..167

         ld             a, [de]         ; a <- X-Position (0-255)

         ld             [hl-], a        ; store out X-Position to COLLMO

         dec            a               ; 0 -> 255, 168..255 -> 167..254
         cp             168 - 1         ; was X-Position = 0 or >= 168 ?
         rla                            ; bit #0 set if TRUE

         and            c
         rrca                           ; was C set and X-Pos in 1..167 ?
         jr             nc, .NoColl     ; * no, so move on to the next MO

         set            7, [hl]         ; make sure CollFlag = TRUE

         ; 3rd, check if any attributes that affect collisions have changed

         inc            e               ; skip X-Position in MO
         inc            e               ; skip Y-Position in MO

         ld             a, [de]         ; a <- MO attribute byte #1
         ld             b, a            ; save a copy in b

         dec            e               ; pos MO @ Y-Position

         sla            c
         sla            c               ; put YM and XM in bits 4-5, C in #2
         xor            c               ; put YM & XM in bits 4-5
         xor            [hl]            ; xor them w/ old CollFlag + MoSeven

         and            $3F             ; YM, XM, YR, XS, YS or YF changed ?
         jr             nz, .Update     ; yes, so update all other fields

         ; attributes are the same, but Y-Position needs updating

         ld             a, [de]         ; a <- Y-Position (0-127)

         inc            l               ; pos COLLMO @ X-Position
         inc            l               ; pos COLLMO @ Width
         inc            l               ; pos COLLMO @ Y-Position

         ld             [hl], a         ; store out Y-Position to COLLMO

         bit            7, b            ; was ReCopy set ?
         jr             nz, .ReCopy_1   ; yes, so update DataAddr

         ; attributes were the same, but ReCopy is NOT set

         ld             a, e
         add            MO_RECSIZE - 2
         ld             e, a            ; pos MO @ next MO's record

         ld             a, l
         add            COLLMO_RECSIZE - 3 + 1
         ld             l, a            ; pos COLLMO @ next MO's record

         cp             ( COLLMO + COLLMO_RECSIZE * 8 + 1 ) & $FF
         jr             nz, .Loop

         ret

.ReCopy_1

         ; attributes were the same, and ReCopy IS set

         inc            e               ; skip Y-Position
         inc            e               ; skip MO attribute byte #1

         ld             a, l
         add            17 - 3
         ld             l, a            ; pos COLLMO @ DataAddr

         cp             ( COLLMO + COLLMO_RECSIZE * 7 + 17 ) & $FF

         ld             a, [de]         ; a <- LOW byte of DataAddr
         inc            e
         ld             [hl+], a        ; store out LOW byte of DataAddr

         ld             a, [de]         ; a <- HIGH byte of DataAddr
         inc            e
         ld             [hl+], a        ; store out HIGH byte of DataAddr

         inc            l               ; skip CollAddr
         inc            l               ; skip BitParam
         inc            l               ; skip CollFlag + MoSeven

         inc            e
         inc            e               ; skip VRAM address

         jr             c, .Loop

         ret

.NoColl

         ; CollFlag will be = FALSE, but update DataAddr if ReCopy set

         res            7, [hl]         ; make sure CollFlag = FALSE

         inc            e               ; skip X-Position in MO
         inc            e               ; skip Y-Position in MO

         ld             a, [de]         ; a <- MO attribute byte #1
         add            a, a            ; is ReCopy set ?
         jr             c, .ReCopy_2    ; yes, so update DataAddr

         ld             a, e
         add            MO_RECSIZE - 3
         ld             e, a            ; pos MO @ next MO's record

         ld             a, l
         add            COLLMO_RECSIZE + 1
         ld             l, a            ; pos COLLMO @ next MO's record

         cp             ( COLLMO + COLLMO_RECSIZE * 8 + 1 ) & $FF
         jr             nz, .Loop

         ret

.ReCopy_2

         ; CollFlag = FALSE, and ReCopy IS set

         inc            e               ; skip attribute byte #1 in MO

         ld             a, l
         add            17 - 0
         ld             l, a            ; pos COLLMO @ DataAddr

         cp             ( COLLMO + COLLMO_RECSIZE * 7 + 17 ) & $FF

         ld             a, [de]         ; a <- LOW byte of DataAddr
         inc            e
         ld             [hl+], a        ; store LOW of DataAddr to COLLMO

         ld             a, [de]         ; a <- HIGH byte of DataAddr
         inc            e
         ld             [hl+], a        ; store HIGH of DataAddr to COLLMO

         inc            l               ; skip CollAddr
         inc            l               ; skip BitParam
         inc            l               ; skip CollFlag + MoSeven

         inc            e
         inc            e               ; skip VRAM address

         jr             c, .Loop

         ret

.Update

         ; attributes were different, so update ALL fields

         xor            [hl]            ; change YM, XM, YR, XS, YS, and YF
         ld             [hl+], a        ; update CollFlag + MoSeven

         inc            l               ; skip X-Position in COLLMO

         add            a, a
         add            a, a            ; put YM, XM, YR, XS, YS, YF in 2-7
         ld             c, a            ; save them in c

         ld             b, AttrTab >> 8 ; bc -> AttrTab[
                                        ;   + YMIRROR << 7 (inverted)
                                        ;   + XMIRROR << 6
                                        ;   + YRES    << 5
                                        ;   + XSIZE   << 4 (inverted)
                                        ;   + YSIZE   << 3
                                        ;   + YFULL   << 2 ]

         ; 2            Width

         ld             a, [bc]         ; a <- Width (8, 16)
         inc            c
         ld             [hl+], a        ; store out Width to COLLMO

         ; 3            Y-Position

         ld             a, [de]         ; a <- Y-Position (0-127)
         inc            e
         ld             [hl+], a        ; store out Y-Position to COLLMO

         ; 4            Height

         ld             a, [bc]         ; a <- Height (4, 8, 16, 32, 64)
         inc            c
         ld             [hl+], a        ; store out Height to COLLMO

         ; 5            SrcOpCode

         ld             a, [bc]         ; a <- OpCode for LD A, [HL+/-]
         inc            c
         ld             [hl+], a        ; store out SrcOpCode to COLLMO

         ; 6            DstOpCode

         ld             a, [bc]         ; a <- OpCode for INC/DEC (D)E
         ld             [hl+], a        ; store out DstOpCode to COLLMO

         inc            b               ; pos AttrTab @ page #1

         ; 7            X-Size Table

         ld             a, [bc]         ; a <- ZShiftTab / ZShiftTabX >> 8
         dec            c
         ld             [hl+], a        ; store out X-Size Table to COLLMO

         ; 8            X-Mirror Table

         ld             a, [bc]         ; a <- ZShiftTab / XMirrorTab >> 8
         dec            c
         ld             [hl+], a        ; store out X-Mirror Table to COLLMO

         ; 9            Entry Point

         ld             a, [bc]         ; a <- EntryPoint
         dec            c
         ld             [hl+], a        ; store out EntryPoint to COLLMO

         inc            b               ; pos AttrTab @ page #2

         ; 10           Loop Point

         ld             a, [bc]         ; a <- LoopPoint
         inc            c
         ld             [hl+], a        ; store out LoopPoint to COLLMO

         ; 11           OpCode #1

         ld             a, [bc]         ; a <- OpCode #1
         inc            c
         ld             [hl+], a        ; store out OpCode #1 to COLLMO

         ; 12           OpCode #2

         ld             a, [bc]         ; a <- OpCode #2
         inc            c
         ld             [hl+], a        ; store out OpCode #2 to COLLMO

         ; 13           OpCode #3

         ld             a, [bc]         ; a <- OpCode #3
         ld             [hl+], a        ; store out OpCode #3 to COLLMO

         ; 14           AttrY1

         ld             b, c            ; save a copy of attributes in b

         ld             a, c
         swap           a
         and            $C0             ; isolate YS & YF in bits 6-7

         ld             [hl+], a        ; store out AttrY1 to COLLMO

         ld             c, a            ; c = LOW byte of RepLineTab

         ; 15           AttrY2

         ld             a, b
         swap           a
         and            $02             ; isolate YR in bit #1

         rl             b
         adc            a, RowStartTab >> 8  ; put YM in bit #0

         ld             [hl+], a        ; store out AttrY2 to COLLMO

         ; 16           RepLine

         ld             b, RepLineTab >> 8  ; bc -> RepLineTab[
                                            ;     YSIZE << 7
                                            ;   + YFULL << 6 ]

         ld             a, [bc]         ; a <- RepLine (1, 2, 4, 8)
         ld             [hl+], a        ; store out RepLine to COLLMO

         ; 17           LOW byte of DataAddr

         inc            e               ; skip MO attribute byte #1

         ld             a, [de]         ; a <- LOW byte of DataAddr
         inc            e
         ld             [hl+], a        ; store LOW of DataAddr to COLLMO

         ; 18           HIGH byte of DataAddr

         ld             a, [de]         ; a <- HIGH byte of DataAddr
         inc            e
         ld             [hl+], a        ; store HIGH of DataAddr to COLLMO

         inc            e
         inc            e               ; skip VRAM address

         inc            l               ; skip CollAddr
         inc            l               ; skip BitParam
         inc            l               ; skip CollFlag + MoSeven

         ld             a, e
         cp             ( MO + MO_RECSIZE * 8 ) & $FF
         jp             nz, .Loop

         ret

; **
; STIC_CollChkMO:  Set the appropriate bits in the STIC's collision
;                  registers ($18-$1F) if any of MOs collide with one
;                  another.
;
; Parameters:      None.
;
; Modifies:        af, bc, hl, de
;

STIC_CollChkMO_ROM:

         ; for the Src, we'll check MO # 0..6

         ld             hl, COLLMO
         ld             de, COLLMO_RECSIZE - 1
CM_SrcChkC:
         ld             a, [hl+]        ; a <- CollFlag & MoSeven
         add            a, a            ; is "CollFlag" (bit #7) set ?
         jr             c, .SrcChkIntr  ; yes, so check if it intersects

         add            hl, de          ; advance to next COLLMO record
         jr             nz, CM_SrcChkC

         ret

.SrcChkIntr

         ; first, make sure we're not using MO # 7 as the Src

         add            a, a            ; was "MoSeven" (bit #6) reset ?
         ret            nc              ; yes, so no Dst MOs left to check

         ; modify code below for checking intersections on X & Y

         ld             a, [hl+]        ; a <- X-Position (1-167)
         ld             [CM_XP + 1], a  ; store it out to "X-Position"

         ld             a, [hl+]        ; a <- Width (8 or 16)
         ld             [CM_WH + 1], a  ; store it out to "Width & Height"

         ld             a, [hl+]        ; a <- Y-Position (0-127)
         ld             [CM_YP + 1], a  ; store it out to "Y-Position"

         ld             a, [hl+]        ; a <- Height (4, 8, 16, 32, 64)
         ld             [CM_WH + 2], a  ; store it out to "Width & Height"

         ; modify PixelPercise directly

         ld             a, [hl+]        ; a <- OpCode for LD A, [HL+/-]
         ld             [PP_SA], a      ; store it out to "Src Advance"

         inc            l               ; skip OpCode for INC/DEC (D)E

         ld             a, [hl+]        ; a <- ZShiftTab / ZShiftTabX >> 8
         ld             [PP_SS + 1], a  ; store it out to "Src Shift"

         ld             a, [hl+]        ; a <- ZShiftTab / XMirrorTab >> 8
         ld             [PP_SX + 1], a  ; store it out to "Src X-Mirror"

         inc            l               ; skip EntryPoint
         inc            l               ; skip LoopPoint
         inc            l               ; skip OpCode #1
         inc            l               ; skip OpCode #2
         inc            l               ; skip OpCode #3

         ; modify pre-PixelPercise preparation code

         ld             a, [hl+]        ; a <- AttrB1 (YS & YF)
         ld             [CM_B1 + 1], a  ; store it out to "Byte #1"

         ld             a, [hl+]        ; a <- AttrB2 (YR & YM)
         ld             [CM_B2 + 1], a  ; store it out to "Byte #2"

         ld             a, [hl+]        ; a <- RepLine (1, 2, 4, 8)
         ld             [PP_SR + 1], a  ; store it out to "Src RepLine"

         ld             a, [hl+]        ; a <- LOW byte of DataAddr
         ld             [CM_DA + 1], a  ; store it out to "Data Address"

         ld             a, [hl+]        ; a <- HIGH byte of DataAddr
         ld             [CM_DA + 2], a  ; store it out to "Data Address"

         ; modify post-PixelPercise code to flag collisions

         ld             a, [hl+]        ; a <- CollAddr ($18-$1F)
         ld             [CM_SC + 1], a  ; store it out to "Src CollAddr"

         ld             a, [hl+]        ; a <- BitParam ($C6-$FE)
         ld             [CM_DB + 1], a  ; store it out to "Dst BitParam"

         ; position HL at next COLLMO here (required if COLLMO is extended)

         push           hl              ; save next Src MO's MOCOLL record

         jr             rCM_DstChkC     ; see if Src intersects w/ any Dst

         ; no intersection detected, so continue to checking remaining MOs

CM_NoY1:
         ld             d, c            ; re-zero d for 16-bit addition
CM_NoY2:
         dec            l               ; pos @ Dst's Y-Position
         dec            l               ; pos @ Dst's Width
CM_NoX1:
         ld             e, COLLMO_RECSIZE - 1
CM_NoX2:
         dec            l               ; pos @ Dst's X-Position
         add            hl, de          ; pos @ next Dst MO's coll. record

         ; for the Dst, we'll check MO # 1..7, 2..7, 3..7, etc.

rCM_DstChkC:

         ld             a, [hl+]        ; a <- CollFlag & MoSeven
         add            a, a            ; is CollFlag (bit #7) set ?
         jr             c, .DstChkIntr  ; yes, so check if it intersects

         add            hl, de          ; advance to next MO record
         jr             nz, rCM_DstChkC

         ; all Dst MO's checked, so proceed to the next Src MO

         pop            hl              ; restore next Src MO's COLLMO record

         jr             CM_SrcChkC      ; continue to check remaining Src MOs

.DstChkIntr

         ; init registers with Src values

rCM_WH:  ld             bc, 0           ; c <- SrcWidth  (8, 16)
                                        ; b <- SrcHeight (4, 8, 16, 32, 64)

         ; ------------------------------------------------------------------
         ;                    X-Intersection Check
         ; ------------------------------------------------------------------

         ld             a, [hl+]        ; a <- DstX
rCM_XP:  sub            0               ; is DstX > SrcX ?
         jr             c, .DstLeft     ; if CY = 1, Dst to the LEFT of Src
         jr             z, .DstEqu      ; if ZF = 1, Dst is EQUAL to Src

         ; DstX > SrcX, so Dst is to the RIGHT of Src

         cp             c               ; is xDistance < SrcWidth ?
         jr             nc, CM_NoX2     ; no, so no intersection on X

         ld             e, a            ; save xDistance in e (1..15)

         rrca                           ; bit #2 is RESET if NdChkLeft = TRUE
         or             c               ; bit #4 is SET if NdChkRight = TRUE

         jr             .IntrYChk       ; intersection on X, now check Y

.DstEqu

         ; Dst X = SrcX, so Dst is EQUAL to Src

         ld             e, a            ; save xDistance in e (0)

         ld             a, c            ; set bit #4 if SrcWidth = 16
         and            [hl]            ; bit #4 is SET if NdChkRight = TRUE

         jr             .IntrYChk       ; intersection on X, now check Y

.DstLeft

         ; DstX < SrcX, so Dst is to the LEFT of Src

         ld             e, a            ; save xDistance in e (-1..-15)

         add            [hl]            ; a = ColumnsLeft (1..15, if valid)
         jr             nc, CM_NoX1     ; if CY = 0, ColumnsLeft < 0
         jr             z, CM_NoX1      ; if ZF = 1, ColumnsLeft = 0

         add            $07             ; set bit #4 if ColumnsLeft > 8
         and            c               ; bit #4 is SET if NdChkRight = TRUE
.IntrYChk
         ld             [CM_NC + 1], a  ; store it out to "NeedCheck"

         inc            l               ; pos @ DstY

         ; ------------------------------------------------------------------
         ;                    Y-Intersection Check
         ; ------------------------------------------------------------------

         ; e = xDistance or -1 * xDistance
         ; d = 0        (DstRowStart)
         ; c = SrcWidth (SrcRowStart)
         ; b = SrcHeight

         ld             a, [hl+]        ; a <- DstY
rCM_YP:  sub            0               ; is DstY < SrcY ?
         jr             c, .DstAbove    ; yes, so Dst is ABOVE Src

         ; DstY >= SrcY, so Dst is BELOW or on LEVEL with Src

         ld             c, a            ; save SrcRowStart in c

         sub            b               ; is yDistance < SrcHeight ?
         jr             nc, CM_NoY2     ; no, so no intersection on Y

         cpl
         inc            a               ; RowCount = SrcHeight - yDistance

         cp             [hl]            ; is RowCount >= DstHeight ?
         jr             c, .LessThan    ; no, so leave RowCount as it is

         ld             a, [hl]         ; RowCount = DstHeight

         jr             .LessThan

.DstAbove

         ; DstY < SrcY, so Dst is ABOVE Src

         cpl
         inc            a               ; a = yDistance (1-63)

         ld             c, d            ; set SrcRowStart = 0
         ld             d, a            ; save DstRowStart in d

         sub            [hl]            ; is yDistance < DstHeight ?
         jr             nc, CM_NoY1     ; no, so no intersection on Y

         cpl
         inc            a               ; RowCount = DstHeight - yDistance

         cp             b               ; is RowCount >= SrcHeight ?
         jr             c, .LessThan    ; no, so leave RowsLeft as it is

         ld             a, b            ; RowCount = SrcHeight
.LessThan
         add            a, a            ; conv. "chunky" rows to "real" rows
         ld             [CM_RC + 1], a  ; store it out to "RowCount"

         ; ------------------------------------------------------------------
         ;      Calculate DstDataAddr & Modify PixelPercise for Dst
         ; ------------------------------------------------------------------

         ; modify PixelPercise directly

         inc            l               ; pos @ Src OpCode
         inc            l               ; pos @ Dst OpCode

         ld             a, [hl+]        ; a <- OpCode for INC/DEC (D)E
         ld             [PP_DA], a      ; store it out to "Dst Advance"

         ld             a, [hl+]        ; a <- ZShiftTab / ZShiftTabX >> 8
         sub            e               ; a = a - xDistance
         ld             [PP_DS + 1], a  ; store it out to "Dst Shift"

         ld             a, [hl+]        ; a <- XMirrorTab / ZShiftTab >> 8
         ld             [PP_DX + 1], a  ; store it out to "Dst X-Mirror"

         inc            l               ; skip EntryPoint
         inc            l               ; skip LoopPoint
         inc            l               ; skip OpCode #1
         inc            l               ; skip OpCode #2
         inc            l               ; skip OpCode #3

         ld             a, [hl+]        ; a <- YSIZE and YFULL in bits 6-7

         or             d               ; put DstRowStart in bits 0-5
         ld             e, a            ; and save them in e

         ld             d, RepLineTab >> 8  ; de -> RepLineTab[
                                            ;     DstYSIZE    << 7
                                            ;   + DstYFULL    << 6
                                            ;   + DstRowStart << 0 ]

         ld             a, [de]         ; a <- DstRepLineInit (1, 2, 4, 6, 8)
         ld             [PP_DI + 1], a  ; store it out to "Dst RepLineInit"

         ; load DE w/ DstDataAddr and position it at DstRowStart

         ld             a, [hl+]        ; a <- YRES and YMIRROR in bits 0-1

         ld             d, a            ; de -> RowStartTab[
                                        ;     DstYRES     << 9
                                        ;   + DstYMIRROR  << 8
                                        ;   + DstYSIZE    << 7
                                        ;   + DstYFULL    << 6
                                        ;   + DstRowStart << 0 ]

         ld             a, [hl+]        ; a <- RepLine (1, 2, 4, 8)
         ld             [PP_DR + 1], a

         ld             a, [de]         ; a <- DstRowStart (0-15)

         add            [hl]
         ld             e, a

         inc            l               ; pos @ HIGH byte of DstDataAddr

         ld             a, [hl+]
         adc            a, 0
         ld             d, a            ; de -> DstData[ DstRowStart ]

         ; modify post-PixelPercise code to flag collisions

         ld             a, [hl+]        ; a <- CollAddr ($18-$1F)
         ld             [CM_DC + 1], a  ; store it out to "Dst CollAddr"

         ld             a, [hl+]        ; a <- BitParam ($C6-$FE)
         ld             [CM_SB + 1], a  ; store it out to "Src BitParam"

         ; position HL at next COLLMO here (required if COLLMO is extended)

         push           hl              ; save next Dst MO's MOCOLL record

         ; ------------------------------------------------------------------
         ;      Calculate SrcDataAddr & Modify PixelPercise for Src
         ; ------------------------------------------------------------------

rCM_B1:  ld             a, 0            ; a <- SrcYSIZE & SrcYFULL in 6-7
         or             c               ; put SrcRowStart in bits 0-5
         ld             l, a            ; and save them in l

         ld             h, RepLineTab >> 8  ; hl <- RepLineTab[
                                            ;     SrcYSIZE    << 7
                                            ;   + SrcYFULL    << 6
                                            ;   + SrcRowStart << 0 ]

         ld             a, [hl]         ; a <- SrcRepLineInit (1, 2, 4, 6, 8)
         ld             [PP_SI + 1], a  ; store it out to "Src RepLineInit"

rCM_B2:  ld             h, 0            ; hl <- RowStartTab[
                                        ;     SrcYRES     << 8
                                        ;   + SrcYMIRROR  << 9
                                        ;   + SrcYSIZE    << 7
                                        ;   + SrcYFULL    << 6
                                        ;   + SrcRowStart << 0 ]

         ld             c, [hl]         ; c <- SrcRowStart (0-15)
         ld             b, 0            ; clear b for 16-bit addition

rCM_DA:  ld             hl, 0
         add            hl, bc          ; hl -> SrcData[ SrcRowStart ]

         ; ------------------------------------------------------------------
         ;             Pre-PixelPercise preparation code
         ; ------------------------------------------------------------------

rCM_RC:  ld             b, 0            ; b <- RowCount (1-64)
rCM_NC:  ld             a, 0            ; a <- NeedCheck

         and            $14             ; is only NdChkLeft = TRUE ?
         jr             z, .ChkLeft     ; yes, so check only the LEFT side

         cp             $14             ; is only NdChkRight = TRUE ?
         jr             z, .ChkRight    ; yes, so check only the RIGHT side

         ; check both the LEFT and RIGHT halves of the XSIZE'd source bitmap

         push           hl              ; save SrcData[ SrcRowStart ]
         push           de              ; save DstData[ DstRowStart ]

         call           PixelPercise    ; does Dst collide with Src ?

         pop            de              ; restore SrcData[ SrcRowStart ]
         pop            hl              ; restore DstData[ DstRowStart ]

         jr             nz, .Collision  ; if ZF = 0, a collision occurred

         ld             a, [CM_RC + 1]
         ld             b, a            ; b <- RowCount (1-64)

.ChkRight

         ; check ONLY the RIGHT half of the XSIZE'd source bitmap

         ld             a, ZShiftTabX >> 8 + 8
         ld             [PP_SS + 1], a  ; modify "Src Shift" for RIGHT half

         ld             a, [PP_DS + 1]
         add            a, 8
         ld             [PP_DS + 1], a  ; modify "Dst Shift" for RIGHT half

         call           PixelPercise    ; does Dst collide with Src ?

         ld             a, ZShiftTabX >> 8
         ld             [PP_SS + 1], a  ; restore "Src Shift" for LEFT half

         jr             nz, .Collision  ; if ZF = 0, a collision occurred
         jr             z, CM_DstNext   ; if ZF = 1, no collision occurred

.ChkLeft

         ; check ONLY the LEFT half of the XSIZE'd source bitmap

         call           PixelPercise    ; does Dst collide with Src ?

         jr             z, CM_DstNext   ; if ZF = 1, no collision occurred

         ; ------------------------------------------------------------------
         ;                         Collision
         ; ------------------------------------------------------------------

.Collision

         ; first, set appropriate collision bit in the Dst MO's STIC register

rCM_DC:  ld             hl, STIC        ; hl -> STIC[ ( $18 + DstMo # ) * 2 ]
rCM_DB:  set            0, [hl]         ; flag that Src collided w/ Dst

rCM_SC:  ld             l, 0            ; hl -> STIC[ ( $18 + SrcMo # ) * 2 ]
rCM_SB:  set            0, [hl]         ; flag that Dst collided w/ Src

CM_DstNext:

         pop            hl              ; restore next Dst MO's COLLMO record

         ld             de, COLLMO_RECSIZE - 1

         jp             CM_DstChkC

STIC_CollChkMO_ROM_End:

; **
; PixelPercise:  Do pixel percise collision checking on Src and Dst,
;                taking into account the XMIRROR, YMIRROR, XSIZE,
;                YSIZE, and YFULL bits of both Src and Dst MOs.
;
; Parameters:    b  = RowCount
;                c  = 1
;                hl = SrcDataAddr
;                de = DstDataAddr
;
;                * PP_SI (Src RepLineInit),  PP_SA (Src Advance),
;                  PP_SX (Src X-Mirror),     PP_SS (Src Shift),
;                  PP_SR (Src RepLine),
;                  PP_DI (Dst RepLineInit),  PP_DA (Dst Advance),
;                  PP_DX (Dst X-Mirror),     PP_DS (Dst Shift),
;                  PP_DR (Dst RepLine)
;
;                * Must all be set in advance.
;
; Notes:         Copied to Low RAM at start-up.
;
;                If ZF = 0, a collision occurred.
;                If ZF = 1, no collision occurred.
;
; Costs:         74 bytes.
;

PixelPercise_ROM:

         ; setup initial parameters to opcodes already in memory

         ld             a, d
         ld             [PP_DD + 1], a      ; PP_DD: ld d, DstDataAddr >> 8
         ld             a, rPP_SI - ( rPP_JR + 2 )
         ld             [PP_JR + 1], a      ; PP_JR: jr nz, rPP_SI

         ld             c, 1            ; set DstRepLine = 1 (1st time only)

         jr             rPP_SA          ; load SrcB, get SrcDataAddr on stack

         ; jumped to after the 1st iteration only from PP_JR

rPP_SI:  ld             l, 0            ; ld l, SrcRepLineInit
rPP_DI:  ld             c, 0            ; ld c, DstRepLineInit

         ld             a, rPP_Loop - ( rPP_JR + 2 )
         ld             [PP_JR + 1], a  ; loop to rPP_Loop from now on

rPP_Loop:

         ; b  = RowCount        d = DstB                   h = SrcB
         ; c  = DstRepLine      e = LOW of DstDataAddr     l = SrcRepLine

         ; advance the source pointer when necessary (according to YS & YF)

         dec            l                 ; dec SrcRepLine
         jr             nz, rPP_SrcNoAdv  ; don't adv. until SrcRepLine = 0

         pop            hl              ; restore SrcDataAddr

rPP_SA:  ld             a, [hl+]        ; a <- SrcB

         push           hl              ; save SrcDataAddr

         ; process byte from Src, flipping it and/or doubling it if necessary

         ld             l, a            ; l = SrcB

rPP_SX:  ld             h, XMirrorTab >> 8
         ld             l, [hl]         ; l <- XMirrorTab/ZShiftTab[ SrcB ]

rPP_SS:  ld             h, ZShiftTabX >> 8
         ld             h, [hl]         ; h <- SrcB's HIGH nybble "doubled"

rPP_SR:  ld             l, 0            ; ld l, SrcRepLine (1, 2, 4 or 8)

         ; ^ 20 cycles

rPP_SrcNoAdv:

         ; advance the dest. pointer when necessary (according to YS & YF)

         dec            c                 ; dec DstRepLine
         jr             nz, rPP_DstNoAdv  ; don't adv. until DstRepLine = 0

rPP_DD:  ld             d, 0            ; ld d, DstDataAddr >> 8

         ld             a, [de]         ; a <- DstB

rPP_DA:  inc            de              ; 16-bit needed only if DstYRES = 1

         ; process byte from Dst, flipping it and/or doubling it if necessary

         ld             c, a            ; c = DstB

         ld             a, d
         ld             [PP_DD + 1], a  ; update HIGH byte of DstDataAddr

         ld             d, b            ; save RowCount
rPP_DX:  ld             b, XMirrorTab >> 8
         ld             a, [bc]
         ld             c, a            ; c <- XMirrorTab/ZShiftTab[ DstB ]

rPP_DS:  ld             b, ZShiftTabX >> 8
         ld             a, [bc]         ; a <- DstB << x, >> x and/or XSIZE'd
         ld             b, d            ; restore RowCount
         ld             d, a            ; d = DstB

rPP_DR:  ld             c, 0            ; ld c, DstRepLine (1, 2, 4 or 8)

         ; ^ 26 cycles

rPP_DstNoAdv:

         ld             a, h            ; a = SrcB
         and            d               ;   = SrcB & DstB
         jr             nz, rPP_Exit    ; if SrcB & DstB <> 0, bits collided

         ; ^ 5/4 cycles

         dec            b               ; dec RowCount
rPP_JR:  jr             nz, rPP_SI      ; jump relative to rPP_SI or rPP_Loop

         ; ^ 4/3 cycles

         ; * 54 cycles per iteration

rPP_Exit:

         pop            hl              ; restore SrcDataAddr

         ret

PixelPercise_ROM_End:

; **
; STIC_CollChkBkg:  Set bit #8 of the appropriate STIC collision
;                   register ($18-$1F) if an MO collides with any
;                   "ON" bit of the background.
;
; Parameters:       None.
;
; Modifies:         af, bc, hl, de
;

STIC_CollChkBkg:

        ld              hl, COLLMO
        ld              de, COLLMO_RECSIZE - 1
CB_MoChkC:
        ld              a, [hl+]        ; a <- CollFlag + MoSeven
        add             a, a            ; is CollFlag (bit #7) set ?
        jr              c, .MoChkIntr   ; yes, so check if it intersects

        add             hl, de          ; advance to next MO record
        jr              nz, CB_MoChkC

        ret

.MoOffScrn

        ; MO was found to be completely off the screen (background cards)

        dec             l               ; pos @ MO's Y-Position
        dec             l               ; pos @ MO's Width
        dec             l               ; pos @ MO's X-Position

        ld              de, COLLMO_RECSIZE - 1
        add             hl, de          ; pos COLLMO @ next MO's record

        jr              CB_MoChkC       ; continue checking remaining MOs

.MoChkIntr

        ; -------------------------------------------------------------------
        ;                Check for LEFT and RIGHT clipping
        ; -------------------------------------------------------------------

        ; calculate CardX (0-19) and CardWidth (1-3), clipping if necessary

        ld              a, [hl]         ; a <- MO's X-Position (1-167)
        and             $07
        ld              c, a            ; c = #/columns to shift RIGHT

        ld              a, [hl+]        ; a <- MO's X-Position (1-167)
        sub             8               ; is MO's X-Position in 1..7 ?
        jr              nc, .ChkRight   ; no, so no LEFT clipping is needed

        ; MO lies partially off-screen to the LEFT

        ld              c, a            ; c = -1 * #/columns to shift LEFT

        xor             a
        ld              [CardXx8], a    ; set CardXx8 = 0

        ld              a, [hl+]        ; a <- MO's Width (8, 16)
        rrca
        rrca
        rrca                            ; divide it by 8

        jr              .LeftShift      ; skip RIGHT clipping (impossible)

.ChkRight

        ld              [CardXx8], a    ; update CardXx8

        ld              e, [hl]         ; e <- MO's Width (8, 16)

        add             e               ; add MO's Width to MoX - 8
        sub             161             ; is it >= 161 ?
        jr              c, .RightShift  ; no, so RIGHT clipping is not needed

        ; MO is either partially or completely off-screen to the RIGHT

        cpl                             ; a = -1 * #/columns to clip at RIGHT
        add             e               ; add it to MO's Width
        ld              e, a            ; save MO's (new) Width in e

.RightShift

        ld              a, c            ; a = -1..-7 for LEFT shifts
                                        ;      0     for NO shift
                                        ;      1..7  for RIGHT shifts
        add             $07             ; add an extra 8 if RIGHT shifting

        add             e               ; add in MO's (new) Width
        rrca
        rrca
        rrca                            ; divide it by 8
        and             $1F             ; a = CardWidth

        inc             l               ; pos @ MO's Y-Position

.LeftShift

        ld              [CardWidth], a  ; update CardWidth (in memory)

        ; -------------------------------------------------------------------
        ;               Check for TOP and BOTTOM clipping
        ; -------------------------------------------------------------------

        ; calculate CardY (0-11) & RowsLeftInit (1-63), clipping if necessary

        ld              a, [hl]         ; a <- MO's Y-Position (0-127)
        and             $07
        ld              b, a            ; b = #/rows to shift DOWN

        ld              a, [hl+]        ; a <- MO's Y-Position (0-127)
        ld              e, [hl]         ; e <- MO's Height (4, 8, 16, 32, 64)

        sub             8               ; is MO's Y-Position in 0..7 ?
        jr              nc, .ChkBottom  ; no, so no TOP clipping is needed

        ; MO is either partially or completely off-screen at the TOP

        ld              d, a            ; save -1 * MoRowStart in d

        add             e               ; add from -1 to -8 to MO's Height

        jr              nc, .MoOffScrn  ; if CY = 0, MO appears off-screen
        jr              z, .MoOffScrn   ; if ZF = 1, MO appears off-screen

        ld              e, a            ; save MO's (new) Height in e

        xor             a
        ld              [CardYx8], a    ; set CardYx8 = 0
        ld              [BP_CS + 1], a  ; set CardRowStart = 0

        sub             d               ; a = 8 - MO's Y-Position
        ld              d, a            ; save MoRowStart in d (1-8)

        ld              b, 8            ; set RowCount = 8

        jr              .UpShift        ; skip bottom clipping (impossible)

.ChkBottom

        ld              [CardYx8], a    ; update CardYx8

        add             e               ; add MO's Height to MoY - 8
        sub             97              ; is it >= 97 ?
        jr              c, .DownShift   ; no, so BOTTOM clipping not needed

        ; MO is either partially or completely off-screen at the BOTTOM

        cpl                             ; a = -1 * #/rows to clip at BOTTOM
        add             e               ; add it to MO's Height

        jr              nc, .MoOffScrn  ; if CY = 0, MO appears off-screen
        jr              z, .MoOffScrn   ; if ZF = 1, MO appears off-screen

        ld              e, a            ; save MO's (new) Height in e

.DownShift

        ; no TOP clipping was needed, so MO will be shifted DOWN

        ld              a, b            ; a = MoShiftY (0-7)
        ld              [BP_CS + 1], a  ; set CardRowStart = MoShiftY (0-7)

        ld              a, 8
        sub             b               ; a = RowCount (8 - MoShiftY)
        ld              b, a            ; save RowCount in b

.UpShift

        ; set initial values for RowCount and RowsLeftInit

        ld              a, e            ; a = MO's (new) Height
        sub             b               ; is it >= RowCount ?
        jr              nc, .UpdateRows ; yes, so leave RowCount as is

        ld              b, e            ; set RowCount = MO's (new) Height
        xor             a               ; set RowsLeftInit = 0
.UpdateRows
        ld              [BP_RI + 1], a  ; update RowsLeftInit (in memory)
        ld              a, b
        ld              [BP_MI + 2], a  ; update RowCount (in memory)

        inc             l               ; skip Height
        inc             l               ; skip Src OpCode
        inc             l               ; skip Dst OpCode

        ; get ZShiftTab or ZShiftTabX according to XSIZE

        ld              a, [hl+]        ; a <- ZShiftTab / ZShiftTabX
        sub             c               ; subtract it by MoShiftX (-7..+7)
        ld              [BP_MS + 1], a  ; store it out to "Mo Shift"

        ; get XMirrorTab or ZShiftTab according to XMIRROR

        ld              a, [hl+]        ; a <- XMirrorTab / ZShiftTab >> 8
        ld              [BP_MX + 1], a  ; store it out to "Mo X-Mirror"

        ; get EntryPoint and LoopPoint according to YSIZE and YFULL

        ld              a, [hl+]        ; a <- EntryPoint (relative offset)
        ld              [BP_EP + 1], a  ; store it out to "Entry Point"
        ld              a, [hl+]        ; a <- LoopPoint (relative offset)
        ld              [BP_LP + 1], a  ; store it out to "Loop Point"

        ; get OpCodes according to YMIRROR, YRES, YSIZE and YFULL

        ld              a, [hl+]        ; a <- OpCode #1
        ld              [BP_TQ + 0], a  ; store it out to "Mo Technique #0"
        ld              a, [hl+]        ; a <- OpCode #2
        ld              [BP_TQ + 1], a  ; store it out to "Mo Technique #1"
        ld              a, [hl+]        ; a <- OpCode #3
        ld              [BP_TQ + 2], a  ; store it out to "Mo Technique #2"

        ld              a, [hl+]        ; a <- AttrB1 (YS & YF)

        or              d               ; put MoStartRow in bits 0-5
        ld              c, a            ; and save them in c

        ld              b, RepLineTab >> 8  ; bc -> RepLineTab[
                                            ;     YSIZE << 7
                                            ;   + YFULL << 6
                                            ;   + MoRowStart ]

        ld              a, [bc]         ; a <- MoRepLineInit (1, 2, 4, 6, 8)
        rrca                            ; convert it to "chunky" rows
        ld              [BP_MI + 1], a  ; store it out to "Mo RepLineInit"

        ld              a, [hl+]        ; a <- AttrB2 (YR & YM)

        ld              b, a            ; bc -> RowStartTab[
                                        ;   + YRES    << 9
                                        ;   + YMIRROR << 8
                                        ;   + YSIZE   << 7
                                        ;   + YFULL   << 6
                                        ;   + MoRowStart ]

        inc             l               ; skip RepLine

        ld              a, [bc]         ; a <- #/starting row in MO's bitmap

        ; get MoDataAddr, add MoRowStart, and store it out to [BP_MD + 1]

        add             [hl]
        ld              [BP_MD + 1], a  ; store LOW of MoData[ MoRowStart ]

        inc             l               ; pos @ HIGH byte of DataAddr

        ld              a, [hl+]
        adc             a, 0
        ld              [BP_MD + 2], a  ; store HIGH of MoData[ MoRowStart ]

        ; get CollAddr and modify BkgPixelPercise

        ld              a, [hl+]        ; a <- MO's CollAddr ($30-$3F)
        ld              [BP_MC + 1], a  ; store it out "Mo CollAddr"

        inc             l               ; skip BitParam

        push            hl              ; save COLLMO @ next MO's record

        ; calculate the initial CARD address using CardX and CardY

        ld              a, [CardYx8]    ; a <- CardYx8

        ld              l, a
        ld              h, CardYx8ToCard >> 8

        ld              a, [CardXx8]    ; a <- CardXx8

        and             $F8             ; make sure bits 0-2 are clear
        add             [hl]
        ld              [BP_HC + 1], a  ; store it out to LOW of H-Card

        inc             h               ; pos @ HIGH byte of CARD address

        ld              a, [hl]
        adc             a, 0
        ld              [BP_HC + 2], a  ; store it out to HIGH of H-Card

        jp              BkgPixelPercise ; *

; ---------------------------------------------------------------------------
;                    .::** BkgPixelPercise_ROM **::.
; ---------------------------------------------------------------------------

BkgPixelPercise_ROM:

        ;   BP_HC (H-Card)              BP_MI (MoRepLineInit & RowCount)
        ;   BP_CS (CardRowStart)        BP_RI (RowsLeftInit)
        ;   BP_MD (MoDataAddress)       BP_EP (EntryPoint)

rBP_HC: ld              hl, 0           ; hl = "horizontal" CARD address

        ld              a, [hl+]
rBP_CS: add             0               ; = add CardRowStart (0-7, no carry)
        ld              d, [hl]
        ld              e, a            ; de <- new CardDataAddr

        ld              a, l            ; a = LOW byte of H-Card address

        ld              [BP_VL + 1], a  ; store out LOW of new V-Card
        add             CARD_RECSIZE - 1
        ld              [BP_HC + 1], a  ; store out LOW of new H-Card

        ld              a, h            ; a = HIGH byte of H-Card address

        ld              [BP_VH + 1], a  ; store out HIGH of new V-Card
        adc             a, 0
        ld              [BP_HC + 2], a  ; store out HIGH of new H-Card

rBP_MD: ld              hl, 0           ; hl = initial value of MoDataAddr
rBP_MI: ld              bc, 0           ; c  = initial value of MoRepLine
                                        ; b  = initial value of RowCount
rBP_RI: ld              a, 0            ; a  = initial value of RowsLeft
        ld              [BP_RL + 1], a  ; set RowsLeft to initial value

rBP_EP: jr              rBP_TQ          ; jump to initial "entry point"

rBP_VLoopDone:

        ; - Decrement CardWidth and end the loop when it reaches zero.
        ; - Increment MoShift by 8 to advance it to the next card column.

        pop             hl              ; take the MoDataAddr off the stack

        ld              hl, BP_MS + 1

        ld              a, [hl]         ; a <- xShiftTab or xShiftTabX
        add             8               ; advance it one card horizontally
        ld              [hl], a         ; update "Mo Shift"

        ld              hl, CardWidth

        dec             [hl]
        jr              nz, rBP_HC

        ; clean up code here (?)

        pop             hl              ; restore COLLMO @ CollFlag + MoSeven

        ld              de, COLLMO_RECSIZE - 1

        jp              CB_MoChkC

rBP_Collision:

        ; collision occurred, so set appropriate bit in MO's coll. register

        ld              a, 1            ; a = #/WRAM bank w/ STIC
        ldh             [cSVBK], a      ; note WRAM bank # for ISRs
        ldh             [rSVBK], a      ; switch to WRAM bank #1

rBP_MC: ld              l, 0
        ld              h, STIC >> 8    ; hl -> STIC[ ( $18 + MO # ) * 2 ]

        inc             l               ; pos @ HIGH byte of coll. register

        set             0, [hl]         ; set "collision w/ bkg" (bit #8)

        pop             hl              ; take the MoDataAddr off the stack
        pop             hl              ; restore COLLMO @ CollFlag + MoSeven

        ld              de, COLLMO_RECSIZE - 1

        ld              a, 3            ; a = #/WRAM bank w/ CARD
        ldh             [cSVBK], a      ; note WRAM bank # for ISRs
        ldh             [rSVBK], a      ; switch to WRAM bank #3

        jp              CB_MoChkC

        ; -------------------------------------------------------------------
        ;                 .::** Vertical Loop **::.
        ; -------------------------------------------------------------------

        ; b  = RowCount
        ; c  = MoRepLine (if MO is single resolution)
        ; de = CardDataAddr
        ; hl = (pointer to transformed) MoB

rBP_SngLoop:                            ; single res. loop (YS << 1 + YF > 0)

        dec             c               ; dec MoRepLine
        jr              nz, rBP_MoNoAdv ; if not 0, don't advance MoDataAddr

rBP_DblLoop:                            ; double res. loop (YS << 1 + YF = 0)

        pop             hl              ; restore MoDataAddr

rBP_TQ: ld              a, [hl+]        ; a = MoB #1
        or              [hl]            ;   = MoB #1 | MoB #2

        inc             hl              ; 16-bit needed only if YRES = 1

        push            hl              ; save MoDataAddr

        ld              l, a                ; l = MoB
rBP_MX: ld              h, XMirrorTab >> 8  ; hl -> XMirror/ZShiftTab[ MoB ]

        ld              l, [hl]             ; l = MoB flipped (or intact)
rBP_MS: ld              h, ZShiftTabX >> 8  ; hl -> xShiftTab(X)[ MoB ]

rBP_MoNoAdv:

        ld              a, [de]         ; a <- CardB
        inc             e               ; inc CardDataAddr

        and             [hl]               ; a = CardB & MoB
        jr              nz, rBP_Collision  ; if ZF = 0, collision occurred

        dec             b               ; dec RowCount
rBP_LP: jr              nz, rBP_DblLoop ; if ZF = 1, collision did not occur

        ;   BP_RL (RowsLeft)
        ;   BP_VL (V-Card LOW)
        ;   BP_VH (V-Card HIGH)

        ; if RowsLeft > 0, calculate a new RowCount and update RowsLeft

rBP_RL: ld              a, 0              ; a = RowsLeft (#/rows left to check)
        or              a                 ; is it = 0 ?
        jr              z, rBP_VLoopDone  ; then the "vertical" loop is done

        ld              b, 8            ; assume we will check 8 more rows
        sub             b               ; subtract it by current RowCount
        jr              nc, .Update     ; if we have enough left, proceed

        add             b               ; restore RowsLeft
        ld              b, a            ; set RowCount = RowsLeft

        xor             a               ; and set RowsLeft = 0 for next time
.Update
        ld              [BP_RL + 1], a  ; update [BP_RL + 1] (RowsLeft)

        ; advance V-Card by one row and get the new CardDataAddr from it

        push            hl              ; save (ptr. to transformed) MoB

rBP_VL: ld              a, 0            ; a <- LOW byte of V-Card
        add             20 * CARD_RECSIZE
        ld              [BP_VL + 1], a  ; update LOW byte of V-Card
        ld              l, a

rBP_VH: ld              a, 0            ; a <- HIGH byte of V-Card
        adc             a, 0            ; set ZF = 0
        ld              [BP_VH + 1], a  ; update HIGH byte of V-Card
        ld              h, a

        ld              a, [hl-]
        ld              e, [hl]
        ld              d, a            ; de <- new CardDataAddr

        pop             hl              ; restore (ptr. to transformed) MoB

        jr              rBP_LP          ; jump to rBP_SngLoop or rBP_DblLoop

BkgPixelPercise_ROM_End:

; **
; STIC_CollChkScrn:  Set bit #9 of the appropriate STIC collision
;                    register ($18-$1F) if an MO collides with the
;                    "edges" of the screen, defined by the frame
;                    located at (-1, -1, 159, 192).
;
; Parameters:        None.
;
; Modifies:          af, bc, hl, de
;

STIC_CollChkScrn:

        ld              hl, COLLMO
.MoRefresh
        ld              de, COLLMO_RECSIZE - 1
.MoChkC
        ld              a, [hl+]        ; a <- CollFlag + MoSeven
        add             a, a            ; is CollFlag set (bit #7) ?
        jr              c, .MoChkIntr   ; yes, so check if it intersects

        add             hl, de          ; advance to next COLLMO record
        jr              nz, .MoChkC

        ret

.MoChkIntr

        ld              c, a            ; save attributes in c

        ; -------------------------------------------------------------------
        ;         Check for CLIPPING on the RIGHT or LEFT side
        ; -------------------------------------------------------------------

        ld              a, [hl+]        ; a <- MO's X-Position (1-167)

        sub             8               ; 1..7 -> 249..255 (unsigned)
        sub             152 - 8         ; is MO's X-Position < 8 or >= 152 ?
        jr              c, .TB_ChkTop   ; no, so no LEFT or RIGHT clipping

        ; clipping may be required on the RIGHT or LEFT side

        rrca                            ; put MoX MSBs in bits 0-2, LSB in #7
        xor             c
        and             $D7
        xor             c               ; put XM and XS in bits 5 & 3

        ld              e, a            ; save them in e

.TB_ChkTop

        inc             l               ; skip Width

        ; ------------------------------------------------------------------
        ;                 TOP and BOTTOM edge detection
        ; ------------------------------------------------------------------

        ; get row number that intersects w/ the TOP edge (if any)

        ld              a, [hl+]        ; a <- MO's Y-Coord. (MoY)
        sub             8               ; is MoY in 0..7 ?
        jr              nc, .TB_ChkBot  ; no, so no TOP intersection

        inc             d               ; toggle YM below (use ODD rows)

        add             [hl]            ; is RowNum < MoHeight ?
        jr              c, .TB_GetRow   ; yes, so intersection w/ TOP edge

        ; MO completely above the TOP edge of the frame (MoHeight must be 4)

        ld              de, COLLMO_RECSIZE - 4
        add             hl, de          ; advance to next COLLMO record

        ld              e, COLLMO_RECSIZE - 1

        jr              .MoChkC         ; continue to check remaining MOs

.TB_ChkBot

        ; get row number that intersects w/ the BOTTOM edge (if any)

        sub             97
        cpl                             ; a = -1 * ( MoY - 8 - 96 )

        cp              [hl]            ; is RowNum < MoHeight ?
        jr              nc, .RL_ChkLeft ; no, so no TOP or BOTTOM intersect

.TB_GetRow

        ld              bc, 14 - 4
        add             hl, bc          ; pos COLLMO @ AttrY1

        ; get row # in MO's _bitmap_ that intersects w/ TOP or BOTTOM edge

        or              [hl]            ; put YS & YF in bits 6-7
        ld              c, a            ; and save them in c

        inc             l               ; pos @ AttrY2

        ld              a, [hl+]        ; a <- AttrY2 (YR & YM)
        xor             d               ; invert YM if TOP edge intersection
        ld              b, a            ; bc -> RowStartTab[
                                        ;     YRES    << 9
                                        ;   + YMIRROR << 8 (inverted if TOP)
                                        ;   + YSIZE   << 7
                                        ;   + YFULL   << 6
                                        ;   + RowNum ]

        ld              a, [bc]         ; a <- RowNum (now in MO's bitmap)

        ; get MoDataAddr, add RowNum and check for a TOP or BOTTOM intersect

        inc             l               ; * skip RepLine

        add             [hl]
        ld              c, a

        inc             l               ; pos @ HIGH byte of MoDataAddr

        ld              a, [hl+]
        adc             a, 0
        ld              b, a            ; bc -> MoData[ RowNum ]

        ld              a, [bc]
        ld              b, a            ; b <- MoB

        ld              d, MoXToBitMask >> 8  ; de -> MoXToBitMask[
                                              ;     MoX (LSB)  << 7
                                              ;   + Right/Left << 6 (0)
                                              ;   + XMIRROR    << 5
                                              ;   + LeftClip   << 4
                                              ;   + XSIZE      << 3 (inverted)
                                              ;   + MoX (MSBs) << 0

        ld              a, [de]         ; a <- BitMask

        and             b               ; is MoB & BitMask <> 0 ?
        jr              z, .NoCollision ; no, so a collision did not occur

.Collision

        ; MO collided with either TOP, BOTTOM, RIGHT or LEFT screen edge

        ld              a, [hl+]        ; a <- MO's CollAddr ($30-$3F)

        ld              c, a
        ld              b, STIC >> 8    ; hl -> STIC[ $18 + MO # * 2 ]

        inc             c               ; pos @ HIGH byte of coll. register

        ld              a, [bc]
        or              $02             ; set "collision w/ screen edges" bit
        ld              [bc], a

        inc             l               ; pos COLLMO @ next MO's record

        jr              .MoRefresh      ; continue checking remaining MOs

.NoCollision

        ; MO intersected the TOP or BOTTOM edge, but no collision occurred

        ld              d, a            ; = ld d, 0

        ld              a, l
        sub             19 - 4
        ld              l, a            ; * pos COLLMO @ MO's Height

.RL_ChkLeft

        ; ------------------------------------------------------------------
        ;                 RIGHT and LEFT edge detection
        ; ------------------------------------------------------------------

        ld              a, COLLMO_RECSIZE - 1
        cp              e               ; any RIGHT or LEFT clipping ?
        jr              nz, .RL_ChkTop  ; yes, so now check for TOP clipping

        ; no TOP / BOTTOM edge collision, and no RIGHT or LEFT intersection

        ld              e, COLLMO_RECSIZE - 4
        add             hl, de          ; pos COLLMO @ next MO's record

        ld              e, a            ; = ld e, COLLMO_RECSIZE - 1

        jr              .MoChkC         ; continue to check remaining MOs

.RL_ChkTop

        dec             l               ; pos @ MO's Y-Position

        ; -------------------------------------------------------------------
        ;          Check for CLIPPING on the TOP or BOTTOM
        ; -------------------------------------------------------------------

        ld              a, [hl+]        ; a <- MO's Y-Position (0..127)

        sub             9               ; is MoY in 0..8 ?
        jr              nc, .RL_ChkBot  ; no, so no TOP clipping is needed

        cpl                             ; a = 8 - MoY (RowStart)
        ld              b, a            ; save RowStart in b in case valid

        cp              [hl]            ; is RowStart < MoHeight ?
        jr              c, .RL_Convert  ; yes, so totally/partially on-screen

.MoNext

        ; MO's bitmap is completely off the screen at the TOP

        ld              e, COLLMO_RECSIZE - 4
        add             hl, de          ; pos COLLMO @ next MO's record

        ld              e, COLLMO_RECSIZE - 1

        jr              .MoChkC

.RL_ChkBot

        ; intersect w/ RIGHT or LEFT edge -- check for clipping at the BOTTOM

        sub             95              ; is MoY >= 104 ?
        jr              nc, .MoNext     ; yes, so MO is completely off-screen

        ld              b, d            ; set RowStart = 0 (no TOP clipping)

        add             [hl]            ; is RowStop < MoHeight ?
        jr              nc, .RL_Convert ; yes, so don't replace RowStop_YM

        ld              d, a            ; save RowStop_YM in d (replacing 0)

.RL_Convert

        ld              a, l
        add             14 - 4
        ld              l, a            ; pos COLLMO @ AttrY1

        ; get values for RowStart_YM & RowCount in the actual MO's _bitmap_

        ld              a, [hl+]        ; a <- AttrY1 (YS & YF)
        or              b               ; put RowStart in bits 0-5
        ld              c, a            ; and save them in c

        xor             b               ; remove RowStart from bits 0-5
        or              d               ; put RowStop in bits 0-5
        ld              d, a            ; and save them in d

        ld              a, [hl+]        ; a <- AttrY2 (YR & YM)

        ld              b, a            ; bc -> RowStartTab[
                                        ;     YRES    << 9
                                        ;   + YMIRROR << 8 (inverted)
                                        ;   + YSIZE   << 7
                                        ;   + YFULL   << 6
                                        ;   + RowStart ]

        ld              a, [bc]         ; a <- RowStart (now in MO's bitmap)

        ld              c, d
        ld              d, a            ; save RowStart in d (0-15)

        ld              a, b
        xor             $01             ; invert YMIRROR (use ODD rows)
        ld              b, a            ; hl -> RowStartTab[
                                        ;     YRES    << 9
                                        ;   + YMIRROR << 8 (inverted)
                                        ;   + YSIZE   << 7
                                        ;   + YFULL   << 6
                                        ;   + RowStop ]

        ld              a, [bc]         ; a <- RowStop  (now in MO's bitmap)

        ld              b, a            ; save RowStop in b (0-15)

        sub             d               ; is RowStop < RowStart ?
        jr              nc, .NoYM       ; no, so YMIRROR was not set

        ld              d, b            ; set RowStart = RowStop

        cpl
        inc             a               ; a = RowStart - RowStop
.NoYM
        inc             a
        ld              b, a            ; b = RowCount (1-16)

        ; accumulate the bitmap data for each row that is be checked

        inc             l               ; * skip RepLine

        ld              a, [hl+]        ; a <- LOW byte of MoDataAddr
        add             d
        ld              c, a

        ld              a, [hl+]        ; a <- HIGH byte of MoDataAddr
        adc             a, 0

        push            hl              ; save COLLMO @ CollAddr

        ld              h, a
        ld              l, c            ; hl -> MoData[ RowStart ]

        xor             a               ; initialize "accumulator" to 0
.Loop
        or              [hl]            ; combine with previous row's data
        inc             hl              ; advance to the next row

        dec             b               ; dec RowCount
        jr              nz, .Loop

        ; now see if the accumulated data has the appropriate bit set

        set             6, e            ; set "Right/Left" (bit #6)

        ld              l, e
        ld              h, MoXToBitMask >> 8  ; hl -> MoXToBitMask[
                                              ;     MoX (LSB)  << 7
                                              ;   + Right/Left << 6 (1)
                                              ;   + XMIRROR    << 5
                                              ;   + LeftClip   << 4
                                              ;   + XSIZE      << 3 (inverted)
                                              ;   + MoX (MSBs) << 0

        and             [hl]            ; is bit set for ANY row checked ?

        pop             hl              ; restore COLLMO @ CollAddr

        jr              nz, .Collision  ; yes, so a collision occurred

        ; MO intersected with RIGHT or LEFT edge, but no collision occurred

        inc             l               ; skip CollAddr
        inc             l               ; skip BitParam

        jp              .MoRefresh      ; continue to check remaining MOs


