; **
; Intellivision for GameBoy Color v1.1
; Copyright(C) 2000,03 By Carl Mueller, Jr.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

HandleControls::

        call            ReadJoy         ; get joypad info

        ; -------------------------------------------------------------------
        ;   Check if Select is newly pressed, released, or being held down
        ; -------------------------------------------------------------------

        ld              d, 0            ; zero d for calls to HandleControl
        ld              c, d            ; = ld c, gcTableOne

        ld              a, [JoyPress]   ; a <- newly pressed buttons

        bit             PADB_SELECT, a  ; did Select just go down ?
        jr              z, .ChkUp       ; no, so check if it went up

        xor             a
        ld              [Select_Used], a  ; set Select_Used = 0
.ChkUp
        ld              a, [JoyRelease] ; a <- newly released buttons

        bit             PADB_SELECT, a  ; did Select just go up ?
        jr              z, .ChkDown     ; no, so check if being held down

        ld              e, gcSelect

        ld              a, [Select_Used]   ; a <- TableOne / TableTwo
        or              a                  ; was Select used in combination ?
        call            z, .HandleControl  ; if no, handle Select
.ChkDown
        ld              a, [JoyDown]    ; a <- JoyDown
        ld              b, a            ; save a copy in b

        bit             PADB_SELECT, b  ; is Select being held down ?
        jr              z, .TableOne    ; no, so use TableOne

        or              $F0
        inc             a               ; are BtnA, BtnB, & Start down too ?
        jp              z, Reset        ; yes, so reset game

        cp              $F8             ; how about just BtnA & BtnB ?
        jr              z, .Pause       ; yes, so pause game w/ 7 & 9 combo

        ld              c, gcNumConstants  ; yes, so use TableTwo

        ld              a, b            ; a = JoyDown

.TableOne

        ; -------------------------------------------------------------------
        ;                  Convert D-Button to an icConstant
        ; -------------------------------------------------------------------

        swap            a
        and             $0F             ; any D-Button bits set ?
        jr              z, .ConvButtons ; no, so skip to converting buttons

        add             a, DbToGc & $FF
        ld              l, a
        ld              a, DbToGc >> 8
        adc             a, d            ; = adc a, 0
        ld              h, a            ; hl -> DbToGc[ JoyDown.4-7 ]

        ld              e, [hl]         ; e <- gcConstant (0-11)

        call            .HandleControl  ; handle D-Button

.ConvButtons

        ; -------------------------------------------------------------------
        ;                Convert button presses to icConstants
        ; -------------------------------------------------------------------

        bit             PADB_A, b       ; is ButtonA being held down ?
        jr              z, .B           ; no, so check for ButtonB

        ld              e, gcButtonA

        call            .HandleControl  ; handle ButtonA
.B
        bit             PADB_B, b       ; is ButtonB being held down ?
        jr              z, .Start       ; no, so check for Start

        ld              e, gcButtonB

        call            .HandleControl  ; handle ButtonB
.Start
        bit             PADB_START, b   ; is Start being held down ?
        jr              z, .GetCodes    ; no, so go get L/R hardware codes

        ld              e, gcStart

        call            .HandleControl  ; handle Start

.GetCodes

        ; -------------------------------------------------------------------
        ;             Look-up hardware codes for LEFT controller
        ; -------------------------------------------------------------------

        ld              hl, HandDelay
        ld              de, HandCode
        ld              bc, $FF << 8 + icNumConstants
.LeftLoop
        ld              a, [hl+]        ; b <- HandDelay[ icConstant ]
        or              a               ; is it = 0 ?
        jr              z, .NoDelayL    ; yes, so not considered pressed

        dec             hl              ; re-pos HandDelay @ icConstant

        dec             a
        ld              [hl+], a        ; dec HandDelay

        ld              a, [de]         ; a <- HandeCode[ icConstant ]
        and             b
        ld              b, a            ; b = HandLeft
.NoDelayL
        inc             de              ; advance HandDelay

        dec             c               ; dec icNumConstants
        jr              nz, .LeftLoop

        ld              a, b
        ld              [SCRATCHPAD_RAM + $0FF * 2], a  ; store out HandLeft

        ; -------------------------------------------------------------------
        ;             Look-up hardware codes for RIGHT controller
        ; -------------------------------------------------------------------

        ld              de, HandCode
        ld              bc, $FF << 8 + icNumConstants
.RightLoop
        ld              a, [hl+]        ; b <- HandDelay[ icConstant ]
        or              a               ; is it = 0 ?
        jr              z, .NoDelayR    ; yes, so not considered pressed

        dec             hl              ; re-pos HandDelay @ icConstant

        dec             a
        ld              [hl+], a        ; dec HandDelay

        ld              a, [de]         ; a <- HandeCode[ icConstant ]
        and             b
        ld              b, a            ; b = HandRight
.NoDelayR
        inc             de              ; advance HandDelay

        dec             c               ; dec icNumConstants
        jr              nz, .RightLoop

        ld              a, b
        ld              [SCRATCHPAD_RAM + $0FE * 2], a  ; store out HandRight

        ret

.Pause

        ld              a, $5A
        ld              [SCRATCHPAD_RAM + $0FF * 2], a  ; store out HandLeft

        ret

.HandleControl

        ; 1st, convert gcConstant to an icConstant using 1 of 2 map tables

        ld              hl, Select_Used

        ld              [hl], c         ; Select_Used <- TableOne / TableTwo

        ld              l, c
        ld              h, GcToIc >> 8

        add             hl, de          ; hl -> GcToIc

        ld              e, [hl]         ; e <- icD0, icD2, etc.

        ; now, refresh HandDelay with a value from HandDelayRefresh

        ld              hl, HandDelayRefresh
        add             hl, de

        ld              a, [hl]         ; a <- HandDelayRefresh (1-7)

        ld              hl, HandDelay
        add             hl, de

        ld              [hl], a         ; HandDelay <- a

        ret

HandCode::

        DB              $FD             ; icD0                EQU     0
        DB              $F9             ; icD1                EQU     1
        DB              $E9             ; icD2                EQU     2
        DB              $EB             ; icD3                EQU     3
        DB              $FB             ; icD4                EQU     4
        DB              $F3             ; icD5                EQU     5
        DB              $E3             ; icD6                EQU     6
        DB              $E7             ; icD7                EQU     7
        DB              $F7             ; icD8                EQU     8
        DB              $F6             ; icD9                EQU     9
        DB              $E6             ; icD10               EQU     10
        DB              $EE             ; icD11               EQU     11
        DB              $FE             ; icD12               EQU     12
        DB              $FC             ; icD13               EQU     13
        DB              $EC             ; icD14               EQU     14
        DB              $ED             ; icD15               EQU     15

        DB              $B7             ; icK0                EQU     16
        DB              $7E             ; icK1                EQU     17
        DB              $BE             ; icK2                EQU     18
        DB              $DE             ; icK3                EQU     19
        DB              $7D             ; icK4                EQU     20
        DB              $BD             ; icK5                EQU     21
        DB              $DD             ; icK6                EQU     22
        DB              $7B             ; icK7                EQU     23
        DB              $BB             ; icK8                EQU     24
        DB              $DB             ; icK9                EQU     25
        DB              $77             ; icKC                EQU     26
        DB              $D7             ; icKE                EQU     27

        DB              $5F             ; icA0                EQU     28
        DB              $9F             ; icA1                EQU     29
        DB              $3F             ; icA2                EQU     30

HandDelayRefresh::

        REPT            2

        DB              1               ; icD0                EQU     0
        DB              1               ; icD1                EQU     1
        DB              1               ; icD2                EQU     2
        DB              1               ; icD3                EQU     3
        DB              1               ; icD4                EQU     4
        DB              1               ; icD5                EQU     5
        DB              1               ; icD6                EQU     6
        DB              1               ; icD7                EQU     7
        DB              1               ; icD8                EQU     8
        DB              1               ; icD9                EQU     9
        DB              1               ; icD10               EQU     10
        DB              1               ; icD11               EQU     11
        DB              1               ; icD12               EQU     12
        DB              1               ; icD13               EQU     13
        DB              1               ; icD14               EQU     14
        DB              1               ; icD15               EQU     15

        DB              7               ; icK0                EQU     16
        DB              7               ; icK1                EQU     17
        DB              7               ; icK2                EQU     18
        DB              7               ; icK3                EQU     19
        DB              7               ; icK4                EQU     20
        DB              7               ; icK5                EQU     21
        DB              7               ; icK6                EQU     22
        DB              7               ; icK7                EQU     23
        DB              7               ; icK8                EQU     24
        DB              7               ; icK9                EQU     25
        DB              7               ; icKC                EQU     26
        DB              7               ; icKE                EQU     27

        DB              1               ; icA0                EQU     28
        DB              1               ; icA1                EQU     29
        DB              1               ; icA2                EQU     30

        ENDR

GcToIc_ROM::

        IF              CONTROLS == ctDefault

        ; -------------------------------------------------------------------
        ;               .::** Default Control Mappings **::.
        ; -------------------------------------------------------------------

        ; with Select released

        DB              icD0            ; gcPadRight          EQU     0
        DB              icD2            ; gcPadUpRight        EQU     1
        DB              icD4            ; gcPadUp             EQU     2
        DB              icD6            ; gcPadUpLeft         EQU     3
        DB              icD8            ; gcPadLeft           EQU     4
        DB              icD10           ; gcPadDownLeft       EQU     5
        DB              icD12           ; gcPadDown           EQU     6
        DB              icD14           ; gcPadDownRight      EQU     7

        DB              icA0            ; gcButtonA           EQU     8
        DB              icA2            ; gcButtonB           EQU     9
        DB              icKE            ; gcSelect            EQU     10
        DB              icA1            ; gcStart             EQU     11

        ; with Select pressed

        DB              icK6            ; gcPadRight          EQU     0
        DB              icK3            ; gcPadUpRight        EQU     1
        DB              icK2            ; gcPadUp             EQU     2
        DB              icK1            ; gcPadUpLeft         EQU     3
        DB              icK4            ; gcPadLeft           EQU     4
        DB              icK7            ; gcPadDownLeft       EQU     5
        DB              icK8            ; gcPadDown           EQU     6
        DB              icK9            ; gcPadDownRight      EQU     7

        DB              icK0            ; gcButtonA           EQU     8
        DB              icK5            ; gcButtonB           EQU     9
        DB              icKE            ; gcSelect            EQU     10
        DB              icKC            ; gcStart             EQU     11

        ELSE

        INCBIN          "GCTOIC.BIN"

        ENDC
