; **
; Intellivision for GameBoy Color v1.1
; Copyright(C) 2000,03 By Carl Mueller, Jr.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

; ---------------------------------------------------------------------------
;                           .::** MACROS **::.
; ---------------------------------------------------------------------------

rtNone         EQU       0              ; no effect
rtOmitSame     EQU       1              ; omit entries like MOVR R0, R0, etc.
rtOmitLast     EQU       2              ; kludge for MOVR (omits MOVR R0, R1)
rtRS_P1        EQU       4              ; 1 is added to R1
rtRM_P1        EQU       4              ; 1 is added to R1
rtRD_P1        EQU       8              ; 1 is added to R2
rtTail         EQU       16             ; must follow w/ RPAIR_TAIL

RETURN: MACRO

        IF              \1 & mpROMB0 != 0

        ld              a, [c]          ; a <- cROMB0 (current ROM bank #)
        ld              [bc], a         ; restore rROMB0

        ENDC

        reti                            ; return, enabling interrupts

        ENDM

RGBSet: MACRO
        DW       ( ( \3 & $3E ) << 9 ) + ( ( \2 & $3E ) << 4 ) + ( \1 >> 1 )
        ENDM

ldhl:   MACRO
        DB              $F8             ; OpCode for LDHL SP, xx
        DB              \2
        ENDM

SETBANK_ROM: MACRO

         ld             a, BANK(\1)
         ldh            [cROMB0], a     ; note new ROM bank # for ISRs
         ld             [rROMB0], a     ; switch to new RAM bank

         ENDM

SETBANK_RAM: MACRO

         ld             a, \1
         ldh            [cSVBK], a      ; note new RAM bank # for ISRs
         ld             [rSVBK], a      ; switch to new RAM bank

         ENDM

; **
; RPAIR_HEAD:  Generates a "jump-entry" table which loads the HIGH byte of
;              the 16-bit register pair specified by the "/1" parameter
;              w/ the LOW byte of R2, and the LOW byte of the same 16-bit
;              register pair w/ the LOW byte of R1 (depending upon the
;              entry point).
;
; Parameters:  \1 = 16-bit register pair (i.e., HL)
;
;              \2 = R1 starting value (i.e., 5)
;              \3 = R1 ending value (i.e., 4)
;              \4 = R2 starting value (i.e., 7)
;              \5 = R2 ending value (i.e., 0)
;
;              \6 = option flags (see rtXXX constants)
;

RPAIR_HEAD: MACRO

           ; get step value and count for each register

           IF              \3 > \2

R1_STEP    SET             1            ; step forward (positive increment)
R1_COUNT   SET             \3 - \2 + 1

           ELSE

R1_STEP    SET             -1           ; step backward (negative increment)
R1_COUNT   SET             \2 - \3 + 1

           ENDC

           IF              \5 > \4

R2_STEP    SET             1            ; step forward (positive increment)
R2_COUNT   SET             \5 - \4 + 1

           ELSE

R2_STEP    SET             -1           ; step backward (negative increment)
R2_COUNT   SET             \4 - \5 + 1

           ENDC

           ; count up the number of entries that will appear in the table

           IF              \6 & rtOmitSame == 0

TAB_COUNT  SET             R1_COUNT * R2_COUNT

           ELSE

TAB_COUNT  SET             0

R1_NUM     SET             \2           ; R1: \2..\3
           REPT            R1_COUNT

R2_NUM     SET             \4           ; R2: \4..\5
           REPT            R2_COUNT

           IF              R1_NUM != R2_NUM
TAB_COUNT  SET             TAB_COUNT+ 1
           ENDC

R2_NUM     SET             ( R2_NUM + R2_STEP ) & $FF
           ENDR

R1_NUM     SET             ( R1_NUM + R1_STEP ) & $FF
           ENDR

           ENDC                         ; (rtOmitSame check)

           IF              \6 & rtOmitLast != 0

TAB_COUNT  SET             TAB_COUNT + -1

           ENDC                         ; (rtOmitLast check)

           ; prepare some final variables for loop below

R1_P1      SET             ( \6 & rtRS_P1 != 0 ) << 0
R2_P1      SET             ( \6 & rtRD_P1 != 0 ) << 8
R1_START   SET             \2
R2_START   SET             \4

FAR_COUNT  SET             0

           ; finally, let's code the "jump-entry" table

           IF              \6 & rtTail != 0
.Head
           ENDC

R1_NUM     SET             \2           ; R1: \2..\3
           REPT            R1_COUNT

R2_NUM     SET             \4           ; R2: \4..\5
           REPT            R2_COUNT

           IF              TAB_COUNT > 0

           IF              ( \6 & rtOmitSame == 0 ) | ( ( \6 & rtOmitSame != 0 ) & ( R1_NUM != R2_NUM ) )

           IF              TAB_COUNT * 5 - 3 - 2 > 127

           ; distance requires a far jump (4 cycles, 3 bytes)

FAR_COUNT  SET             FAR_COUNT + 1

           IF              \6 & rtTail == 0

           ld              \1, R2_NUM << 9 + R2_P1 + R1_NUM << 1 + R1_P1
           jp              .Start

           ENDC                         ; (rtTail)

           ELSE

           ; distance requires a relative jump (3 cycles, 2 bytes)

           ld              \1, R2_NUM << 9 + R2_P1 + R1_NUM << 1 + R1_P1

           IF              ! ( TAB_COUNT == 1 )
           jr              .Start
           ENDC

           ENDC                         ; (Distance)

TAB_COUNT  SET             TAB_COUNT + -1

           ENDC                         ; (rtOmitSame)

           ENDC                         ; (rtOmitLast)

R2_NUM     SET             ( R2_NUM + R2_STEP ) & $FF
           ENDR

R1_NUM     SET             ( R1_NUM + R1_STEP ) & $FF
           ENDR

.Start

           ENDM

RPAIR_TAIL: MACRO

.Tail

TAB_COUNT  SET             1

R1_NUM     SET             R1_START     ; R1: \2..\3
           REPT            R1_COUNT

R2_NUM     SET             R2_START     ; R2: \4..\5
           REPT            R2_COUNT

           IF              FAR_COUNT > 0

           IF              ( \2 & rtOmitSame == 0 ) | ( ( \2 & rtOmitSame != 0 ) & ( R1_NUM != R2_NUM ) )

           ld              \1, R2_NUM << 9 + R2_P1 + R1_NUM << 1 + R1_P1

           IF              .Tail - .Start + TAB_COUNT * 5 > 128
           jp              .Start       ; far jump (4 cycles, 3 bytes)
           ELSE
           jr              .Start       ; relative jump (3 cycles, 2 bytes)
           ENDC

TAB_COUNT  SET             TAB_COUNT + 1
FAR_COUNT  SET             FAR_COUNT + -1

           ENDC                         ; (rtOmitSame)

           ENDC                         ; (Far_Count)

R2_NUM     SET             ( R2_NUM + R2_STEP ) & $FF
           ENDR

R1_NUM     SET             ( R1_NUM + R1_STEP ) & $FF
           ENDR

           ; end-of-tail marker (required by BIN2IGB)

           DB              ( .Tail - .Head ) >> 8
           DB              ( .Tail - .Head ) & $FF

           ENDM

; **
; RSING_TAB:   Generates a "jump-entry" table which loads the 8-bit
;              register specified by the "/1" parameter w/ the LOW
;              byte of R1 (depending upon the entry point).
;
; Parameters:  \1 = 8-bit register (i.e., A, E, or L)
;
;              \2 = R1 starting value (i.e., 7)
;              \3 = R1 ending value (i.e., 0)
;
;              \4 = option flags (see rtXXX constants)
;

RSING_TAB: MACRO

           ; get step value and count

           IF              \3 > \2

R1_STEP    SET             1            ; step forward (positive increment)
R1_COUNT   SET             \3 - \2 + 1

           ELSE

R1_STEP    SET             -1           ; step backward (negative increment)
R1_COUNT   SET             \2 - \3 + 1

           ENDC

           ; finally, let's code the "jump-entry" table

R1_NUM     SET             \2           ; R1: \2..\3
           REPT            R1_COUNT

           IF              \4 & ( rtRS_P1 | rtRD_P1 ) != 0
           ld              \1, R1_NUM << 1 + 1
           ELSE
           ld              \1, R1_NUM << 1
           ENDC

           IF              R1_NUM != \3
           jr              .Start
           ENDC

R1_NUM     SET             ( R1_NUM + R1_STEP ) & $FF
           ENDR

.Start

           ENDM

; **
; JUMP:        Converts the CP1610 address passed into a "real" address,
;              loads it into [R7], switches to the appropriate bank, and
;              updates cROMB0 (for ISRs) and R7_Conv (to convert back
;              to CP1610 when necessary).
;
; Parameters:  h  = AddrHiToROMB0 >> 8 ( = R0 >> 8 )
;              l  = HIGH byte of Jump Address (CP1610)
;              e  = LOW byte of Jump Address (CP1610)    <- * IDEAL
;
;              If call was made to Jump_Op,
;
;                 a = HIGH byte of CP1610 Jump address
;                 e = LOW byte of CP1610 Jump address
;
;              If call was made to Jump_Peek,
;
;                 d = HIGH byte of CP1610 Jump address
;                 a = LOW byte of CP1610 Jump address
;
; Modifies:    a, hl, sp
;

JUMP:   MACRO

        IF              \1 & mpR0inH == 0

        ld              h, c            ; h = R0 >> 8

        ENDC

        IF              \1 & mpJAinLA != 0

        ld              e, a            ; e = LOW of Jump Addr. (CP1610)

        ELSE

        ld              l, a            ; l = HIGH of Jump Addr. (CP1610)

        ENDC

        ld              a, [hl]         ; a <- ROM bank #
        ld              [c], a          ; update cROMB0
        ld              [bc], a         ; update rROMB0

        inc             h               ; pos @ AddrHiToReal >> 8

        ld              h, [hl]         ; h <- HIGH of Jump Addr. (real) >> 1

        ld              a, l            ; a = HIGH of Jump Addr. (CP1610)
        sub             h
        ldh             [R7_Conv], a

        ld              l, e            ; l = LOW of Jump Addr. (CP1610)

        add             hl, hl
        ld              sp, hl          ; sp = Jump Address (real)

        RETURN          0               ; return, enabling interrupts

        ; ^ 24-25 cycles, 14-15 bytes

        ENDM

; **
; PEEK:        Gets the word stored at a memory location within the
;              CP1610's address space.  If \1 == 1 (indicating that
;              SDBD is set), this word is formed from the LOW bytes
;              of two consecutive addresses.
;
; Parameters:  a = HIGH byte of address to peek
;              d = LOW byte of address to peek
;              h = R0 >> 8
;
; Costs:       26 bytes.
;

PEEK:   MACRO

        ; Peek to:      STIC            $0000-$003F -> $DC00-$DC7F
        ;               SCRATCHPAD RAM  $0100-$01FF -> $D600-$D7FF
        ;               SYSTEM RAM      $0200-$035F -> $D000-$D2BF

        ; Peek to:      EXEC            $1000-$1FFF -> $4000-$5FFF (2)
        ;               GROM_16         $3000-$37FF -> $6000-$6FFF (2)
        ;               GRAM_16         $3800-$39FF -> $DC00-$DFFF
        ;               GAME_1          $5000-$6FFF -> $4000-$7FFF (3)
        ;               GAME_2          $D000-$DFFF -> $4000-$5FFF (4)
        ;                               $F000-$FFFF -> $6000-$7FFF (4)

        ld              l, a            ; hl -> AddrHiToROMB0[ a ]

        ld              a, [hl]         ; a <- ROM bank # where binary is
        ld              [bc], a         ; switch to appropriate ROM bank

        inc             h               ; hl -> AddrHiToReal[ a ]

        ld              h, [hl]         ; h <- HIGH byte of address (conv'd)
        ld              l, d            ; l = LOW byte of address to peek

        add             hl, hl          ; hl -> 16-bit value to be peeked

        IF              \1 & mpSDBD == 0  ; SDBD = 0

        ld              a, [hl+]        ; a  <- LOW byte of 16-bit value
        ld              \2, [hl]        ; \2 <- HIGH byte of 16-bit value

        ELSE                              ; SDBD = 1

        ld              a, [hl+]        ; a  <- LOW byte of 16-bit value

        IF              \1 & mpSDBD_R13 != mpSDBD_R13

        inc             hl
        ld              \2, [hl]        ; \2 <- HIGH byte of 16-bit value

        ENDC

        ENDC

        ; ^ 15 cycles, 9 bytes

        ENDM

; **
; POKE:        Poke a word in the CP1610 address space.
;
; Parameters:  a  = LOW byte of ADDR
;              h  = HIGH byte of ADDR
;              de = (pointer to) RS
;

POKE:   MACRO

        ld              l, c                  ; l = R0 >> 8 ($D3)

        jp              [hl]                  ; jump to "poke" handler

        ; ^ 2 cycles, 2 bytes (!)

        ENDM

; ---------------------------------------------------------------------------
;       .::** MACROS used by implementing functions for the CP1610 **::.
; ---------------------------------------------------------------------------

; **
; ADD_RD:      Add DA to the destination register pointed to by HL.
;
; Parameters:  da = (16-bit) value to be added
;              hl = (pointer to) RD
;
; Modifies:    a  = HIGH byte of Result
;              hl = (pointer to) RD + 1
;

ADD_RD: MACRO

        add             [hl]            ; a = LOW byte of Dst + Src
        ld              [hl+], a        ; store out LOW byte of Dst + Src

        ld              a, d            ; a <- HIGH byte of Src

        adc             [hl]            ; a = HIGH byte of Dst + Src
        ld              [hl], a         ; store out HIGH byte of Dst + Src

        RETURN          \1              ; return, enabling interrupts

        ; ^ 13 cycles, 6 bytes

        ENDM

; **
; ADD_RD_F:    Add DA to the destination register pointed to by HL, setting
;              the Sign, Zero, Overflow and Carry flags accordingly.
;
; Parameters:  da = (16-bit) value to be added
;              hl = (pointer to) RD
;
; Modifies:    a  = new Status word (SZOC in bits 3..0)
;              e  = HIGH byte of [RD] before it was modified
;              hl = (pointer to) RD + 1
;

ADD_RD_F: MACRO

        add             [hl]            ; a = LOW byte of [RD] + Src
        ld              [hl+], a        ; store out LOW byte of [RD] + Src

        ld              a, [hl]         ; a <- HIGH byte of [RD]
        ld              e, a            ; save a copy in e

        jr              z, .ChkZero     ; branch if LOW byte of Result = 0

        adc             a, d            ; a = HIGH byte of [RD] + Src
        ld              [hl], a         ; store out HIGH byte of [RD] + Src

.ChkCarry

        jr              c, .Carry

.NoCarry

        ; - Sign is set only if bit #15 of Result is set.
        ; x Zero is NOT set.
        ; - Overflow is set only if both Src and Dst are positive and
        ;   the result is negative.
        ; x Carry flag is NOT set (so Src and Dst can't both be negative).
        ;
        ;   S Z O C
        ;   ? 0 ? 0

        add             a, a            ; is sign bit of Result set ?
        jr              c, .ChkOver_1   ; yes, so an overflow may be possible

        ;   S Z O C
        ;   0 0 0 0

        ld              b, $20          ; clear all flags

        RETURN          \1              ; return, enabling interrupts

.ChkOver_1

        ld              a, e
        or              d               ; bit #7 = 0 if Src & Dst were both +

        add             a, a            ; is Overflow (bit #7) reset ?
        jr              nc, .Over_1

        ;   S Z O C
        ;   1 0 0 0

        ld              b, $28          ; set Sign only

        RETURN          \1              ; return, enabling interrupts

.Over_1

        ;   S Z O C
        ;   1 0 1 0

        ld              b, $2A          ; set Sign and Overflow

        RETURN          \1              ; return, enabling interrupts

.Carry

        ; - Sign is set only if bit #15 of Result is set.
        ; x Zero is NOT set.
        ; - Overflow is set only if both Src and Dst are negative and
        ;   the result is positive.
        ; x Carry flag IS set (so Src and Dst can't both be positive).
        ;
        ;   S Z O C
        ;   ? 0 ? 1

        add             a, a            ; is bit #15 of Result set ?
        jr              c, .NoOver_2

        ld              a, e
        and             d               ; bit #7 = 1 if Src & Dst were both -

        add             a, a            ; is Overflow (bit #7) set ?
        jr              c, .Over_2

        ;   S Z O C
        ;   0 0 0 1

        ld              b, $21          ; set Carry only

        RETURN          \1              ; return, enabling interrupts

.Over_2

        ;   S Z O C
        ;   0 0 1 1

        ld              b, $23          ; set Overflow and Carry

        RETURN          \1              ; return, enabling interrupts

.NoOver_2

        ;   S Z O C
        ;   1 0 0 1

        ld              b, $29          ; set Sign and Carry

        RETURN          \1              ; return, enabling interrupts

.ChkZero

        adc             a, d            ; a = HIGH byte of [RD] + Src
        ld              [hl], a         ; store out HIGH byte of [RD] + Src

        jr              nz, .ChkCarry     ; if Z = 0, HIGH of Src + Dst <> 0
        jr              c, .ZeroAndCarry  ; if C = 1, Src = X and Dst = -X

        ;   S Z O C
        ;   0 1 0 0

        ld              b, $24          ; set Zero only

        RETURN          \1              ; return, enabling interrupts

.ZeroAndCarry

        ; x Sign flag is NOT set.
        ; x Zero flag IS set.
        ; - Overflow is set only if both the Src and Dst were = -32,768.
        ; x Carry IS set.
        ;
        ;   S Z O C
        ;   0 1 ? 1

        rra                             ; a = $80
        cp              d               ; is HIGH byte of Src = $80 ?
        jr              z, .Over_3      ; yes, so overflow occurred

        ld              b, $25          ; set Zero and Carry

        RETURN          \1              ; return, enabling interrupts

.Over_3

        ld              b, $27          ; set Zero, OverFlow, and Carry

        RETURN          \1              ; return, enabling interrupts

        ENDM

; **
; ADD_RD_8:    Add A to the 16-bit destination register pointed to by HL.
;
; Parameters:  a  = (8-bit) value to be added
;              hl = (pointer to) RD
;
; Modifies:    a  = LOW byte of Result
;              hl = (pointer to) RD + 1
;

ADD_RD_8: MACRO

        add             [hl]            ; a = LOW byte of Dst + Src
        ld              [hl+], a        ; store out LOW byte of Dst + Src

        IF              \1 & mpROMB0 != 0

        ld              a, [c]          ; a <- cROMB0 (current ROM bank #)
        ld              [bc], a         ; restore rROMB0

        ENDC

        ret             nc              ; *

        inc             [hl]            ; increment HIGH byte of Dst

        reti                            ; return, enabling interrupts

        ENDM

; **
; ADD_RD_8_F:  Add A to the 16-bit destination register pointed to by HL,
;              setting the Sign, Zero, Overflow and Carry flags accordingly.
;
; Parameters:  a  = 8-bit value to be added
;              hl = (pointer to) RD
;

ADD_RD_8_F: MACRO

        add             [hl]            ; a = LOW byte of [RD] + Src
        jr              z, .ChkZero     ; branch if LOW byte of Result = 0

        ld              [hl+], a        ; store out LOW byte of [RD] + Src

        jr              c, .Carry

        ld              a, [hl]         ; a <- HIGH byte of [RD]

        ; - Sign is set only if bit #15 of Dst is set.
        ; x Zero is NOT set.
        ; x Overflow is NOT set.
        ; x Carry flag is NOT set.

.ChkSign

        ;   S Z O C
        ;   ? 0 0 0

        add             a, a            ; is Sign (bit #7) set ?
        jr              c, .Sign

.ClearAll

        ld              b, $20          ; clear all flags

        RETURN          \1              ; return, enabling interrupts

.ChkZero

        jr              nc, .GetSignAndZero  ; Src & LOW byte of Dst are = 0

        ; NOTE:  Addition of the 8-bit Src and the LOW byte of 16-bit Dst
        ;        was zero but produced a carry (the positive and negative
        ;        versions of the same number were added together), so the
        ;        HIGH byte of the Dst must be incremented.

        ld              [hl+], a        ; store out LOW byte of [RD] + Src

        inc             [hl]            ; inc HIGH byte of Dst
        jr              nz, .GetSignAndOverFlow

        ;   S Z O C
        ;   0 1 0 1

        ld              b, $25          ; set Zero and Carry

        RETURN          \1              ; return, enabling interrupts

.GetSignAndZero

        ;   S Z O C
        ;   ? ? 0 0

        inc             l               ; pos @ HIGH byte of Dst

        or              [hl]            ; is HIGH byte of Dst = 0 too ?
        jr              nz, .ChkSign    ; no, so just check for Sign

        ;   S Z O C
        ;   0 1 0 0

        ld              b, $24          ; set Zero only

        RETURN          \1              ; return, enabling interrupts

.Carry

        ; NOTE:  Addition of 8-bit Src and the LOW byte of 16-bit Dst
        ;        produced a carry, so the HIGH byte of Dst must be
        ;        incremented.

        inc             [hl]            ; inc HIGH byte of Dst
        jr              z, .SetCarry    ; HIGH byte of Dst was = $FF

.GetSignAndOverFlow

        ; - Sign is set only if bit #15 of Result is set.
        ; x Zero is NOT set.
        ; - Overflow is set only if both Src and Dst are positive and
        ;   the result is negative.
        ; x Carry flag is NOT set (so Src and Dst can't both be negative).
        ;
        ;   S Z O C
        ;   ? 0 ? 0

        ld              a, [hl]         ; a <- HIGH byte of Dst
        add             a, a            ; is sign (bit #7) set ?
        jr              nc, .ClearAll   ; no, so no overflow is possible

        ;   S Z O C
        ;   1 0 ? 0

        jr              z, .OverFlow    ; HIGH byte of Result = $80

.Sign

        ;   S Z O C
        ;   1 0 0 0

        ld              b, $28          ; set Sign only

        RETURN          \1              ; return, enabling interrupts

.OverFlow

        ;   S Z O C
        ;   1 0 1 0

        ld              b, $2A          ; set Sign and Overflow

        RETURN          \1              ; return, enabling interrupts

.SetCarry

        ;   S Z O C
        ;   0 0 0 1

        ld              b, $21          ; set Carry Only

        RETURN          \1              ; return, enabling interrupts

        ENDM

; **
; SUB_RD:      Subtract DE from the destination register pointed to by HL.
;
; Parameters:  de = (16-bit) value to be subtracted
;              hl = (pointer to) RD
;
; Modifies:    a  = HIGH byte of Result
;              hl = (pointer to) RD + 1
;

SUB_RD: MACRO

        ld              a, [hl]         ; a <- LOW byte of Dst
        sub             e               ; a = LOW byte of Dst - Src
        ld              [hl+], a        ; store out LOW byte of Dst - Src

        ld              a, [hl]         ; a <- HIGH byte of Dst
        sbc             d               ; a = HIGH byte of Dst - Src
        ld              [hl], a         ; store out HIGH byte of Dst - Src

        RETURN          \1              ; return, enabling interrupts

        ; ^ 14 cycles, 7 bytes

        ENDM

; **
; SUB_RD_F:    Subtract DE from the destination register pointed to by HL,
;              setting the Sign, Zero, Overflow, and Carry flags accordingly.
;
; Parameters:  de = value to be subtracted
;              hl = (pointer to) RD
;
; Modifies:    a  = new Status word (SZOC in bits 3..0)
;              e  = HIGH byte of [RD] before it was modified
;

SUB_RD_F: MACRO

        IF              \1 & mpCmp == 0 ; store result (subtraction)

        ld              a, [hl]         ; a <- LOW byte of Dst
        sub             e               ; a = LOW byte of Dst - Src
        ld              [hl+], a        ; store out LOW byte of Dst - Src

        ld              a, [hl]         ; a <- HIGH byte of Dst
        ld              e, a            ; save a copy in e

        jr              z, .ChkZero

        sbc             d               ; a = HIGH byte of Dst - Src
        ld              [hl], a         ; store out HIGH byte of Dst - Src

        ELSE                            ; throw away result (comparison)

        ld              a, [hl+]        ; a <- LOW byte of Dst
        sub             e               ; a = LOW byte of Dst - Src

        ld              a, [hl]         ; a <- HIGH byte of Dst
        ld              e, a            ; save a copy in e

        jr              z, .ChkZero

        sbc             d               ; a = HIGH byte of Dst - Src

        ENDC

        jr              c, .Carry

.NoCarry

        ; - Sign is set only if bit #15 of the Result is set.
        ; x Zero is NOT set.
        ; - Overflow is set only if the signs of the Src and Dst differ
        ;   (positive and negative, respectively) and the Result is positive.
        ; x Carry IS set (because it is inverted on the CP1610 for
        ;   subtraction operations).
        ;
        ;   S Z O C
        ;   ? 0 ? 1

        add             a, a            ; is Sign (bit #7) set ?
        jr              c, .NoOver_1    ; yes, so an overflow is not possible

        ld              a, e            ; a = HIGH byte of Dst
        xor             d               ;   = HIGH byte of Dst ^ Src

        add             a, a            ; is Overflow (bit #7) set ?
        jr              c, .Over_1

        ;   S Z O C
        ;   0 0 0 1

        ld              b, $21          ; set Carry Only

        RETURN          \1              ; return, enabling interrupts

.Over_1

        ;   S Z O C
        ;   0 0 1 1

        ld              b, $23          ; set Overflow and Carry

        RETURN          \1              ; return, enabling interrupts

.NoOver_1

        ;   S Z O C
        ;   1 0 0 1

        ld              b, $29          ; set Sign and Carry

        RETURN          \1              ; return, enabling interrupts

.ChkZero

        sbc             d               ; a = HIGH byte of Dst - Src

        IF              \1 & mpCmp == 0 ; store result (subtraction)

        ld              [hl], a         ; store out HIGH byte of Dst - Src

        ENDC

        jr              z, .Zero
        jr              nc, .NoCarry

.Carry

        ; - Sign is set only if bit #15 of the Result is set.
        ; x Zero is NOT set.
        ; - Overflow is set only if the signs of the Src and Dst differ
        ;   (negative and positive, respectively) and the Result is negative.
        ; x Carry is NOT set (because it is inverted on the CP1610 for
        ;   subtraction operations).
        ;
        ;   S Z O C
        ;   ? 0 ? 0

        add             a, a            ; is bit #7 set ?
        jr              nc, .NoOver_2   ; no, so an overflow is not possible

        ld              a, e            ; a = HIGH byte of Dst
        xor             d               ;   = HIGH byte of Dst ^ Src

        add             a, a            ; is bit #7 set ?
        jr              c, .Over_2      ; yes, so an overflow occurred

        ;   S Z O C
        ;   1 0 0 0

        ld              b, $28          ; set Sign

        RETURN          \1              ; return, enabling interrupts

.Over_2

        ;   S Z O C
        ;   1 0 1 0

        ld              b, $2A          ; set Sign and Overflow

        RETURN          \1              ; return, enabling interrupts

.NoOver_2

        ;   S Z O C
        ;   0 0 0 0

        ld              b, $20          ; clear all flags

        RETURN          \1              ; return, enabling interrupts

.Zero

        ;   S Z O C
        ;   0 1 0 1

        ld              b, $25          ; set Zero and Carry

        RETURN          \1              ; return, enabling interrupts

        ENDM

; **
; SUB_RD_8:    Subtract E from the destination register pointed to by HL.
;
; Parameters:  e  = (8-bit) value to be subtracted
;              hl = (pointer to) RD
;
; Modifies:    a, hl
;

SUB_RD_8: MACRO

        ld              a, [hl]         ; a <- LOW byte of Dst
        sub             e               ; a = LOW byte of Dst - Src
        ld              [hl+], a        ; store out LOW byte of Dst - Src

        IF              \1 & mpROMB0 != 0

        ld              a, [c]          ; a <- cROMB0 (current ROM bank #)
        ld              [bc], a         ; restore rROMB0

        ENDC

        ret             nc

        dec             [hl]            ; dec HIGH byte of Dst

        reti                            ; return, enabling interrupts

        ENDM

; **
; SUB_RD_8_F:  Subtract E from the 16-bit destination register pointed to
;              by HL, setting the Sign, Zero, Overflow, and Carry flags
;              accordingly.
;
; Parameters:  e  = (8-bit) value to be subtracted
;              hl = (pointer to) RD
;
; Modifies:    a  = new Status word (SZOC in bits 3..0)
;

SUB_RD_8_F: MACRO

        IF              \1 & mpCmp == 0 ; store result (subtraction)

        ld              a, [hl]         ; a <- LOW byte of Dst
        sub             e               ; a = LOW byte of Dst - Src
        ld              [hl+], a        ; store out LOW byte of Dst - Src

        ELSE                            ; discard result (comparison)

        ld              a, [hl+]        ; a <- LOW byte of Dst
        sub             e

        ENDC

        jr              z, .ChkZero     ; \

        ld              a, [hl]         ; a <- HIGH byte of Dst

        jr              nc, .GetSign    ; / HIGH of Dst need not be dec'd

        ; NOTE: Subtraction of 8-bit Src from the LOW byte of 16-bit Dst
        ;       produced a carry, so the HIGH byte of the Dst MUST be
        ;       decremented.

        IF              \1 & mpCmp == 0 ; store result (subtraction)

        dec             [hl]            ; decrement HIGH byte of Dst

        ENDC

        or              a               ; was the HIGH byte of Dst = 0 ?
        jr              nz, .NoCarry    ; no, so no carry was produced

        ;   S Z O C
        ;   1 0 0 0

        ld              b, $28          ; set Sign

        RETURN          \1              ; return, enabling interrupts

.NoCarry

        ; - Sign is set only if bit #15 of the Result is set.
        ; x Zero is NOT set.
        ; - Overflow is set only if the signs of the Src and Dst differ
        ;   (positive and negative, respectively) and the Result is positive.
        ; x Carry IS set (because it is inverted on the CP1610 for
        ;   subtraction operations).
        ;
        ;   S Z O C
        ;   ? 0 ? 1

        add             a, a            ; was sign (bit #7) set ?
        jr              nc, .NoSign     ; no, so possibility of OverFlow
        jr              z, .OverFlow    ; jump to OverFlow if Dst was = $80

.Sign

        ;   S Z O C
        ;   1 0 0 1

        ld              b, $29          ; set Sign and Carry (inverted)

        RETURN          \1              ; return, enabling interrupts

.OverFlow

        ;   S Z O C
        ;   0 0 1 1

        ld              b, $23          ; set Overflow and Carry (inverted)

        RETURN          \1              ; return, enabling interrupts

.ChkZero

        ; NOTE: Subtraction of 8-bit Src from the LOW byte of 16-bit Dst
        ;       produced a zero, so it is impossible for the carry to be
        ;       set.  Consequently, the HIGH byte of the Dst will NOT be
        ;       decremented.

        ; - Sign is set only if bit #15 of the Result is set.
        ; - Zero is set only if HIGH byte of the Result is zero.
        ; x Overflow is NOT set.
        ; x Carry IS set (because it is inverted on the CP1610 for
        ;   subtraction operations).

        or              [hl]            ; is HIGH byte of Dst = 0 too ?
        jr              nz, .GetSign

        ;   S Z O C
        ;   0 1 0 1

        ld              b, $25          ; set Zero and Carry (inverted)

        RETURN          \1              ; return, enabling interrupts

.GetSign

        ;   S Z O C
        ;   ? 0 0 1

        add             a, a            ; is Sign (bit #7) set ?
        jr              c, .Sign

.NoSign

        ;   S Z O C
        ;   0 0 0 1

        ld              b, $21          ; set Carry

        RETURN          \1              ; return, enabling interrupts

        ENDM

; **
; AND_RD:      AND [RD] w/ DA.
;
; Parameters:  da = value to AND w/ [RD]
;              hl = (pointer to) RD
;

AND_RD: MACRO

        and             [hl]            ; a = LOW byte of Dst & Src
        ld              [hl+], a        ; store out LOW byte of Dst & Src

        ld              a, d            ; a <- HIGH byte of Src

        and             [hl]            ; a = HIGH byte of Dst & Src
        ld              [hl], a         ; store out HIGH byte of Dst & Src

        RETURN          \1              ; return, enabling interrupts

        ; ^ 13 cycles, 6 bytes

        ENDM

; **
; AND_RD_F:    AND [RD] with da, setting the Sign and Zero flags
;              accordingly.
;
; Parameters:  da = value to AND w/ [RD]
;              hl = (pointer to) RD
;

AND_RD_F: MACRO

        and             [hl]            ; a = LOW byte of Dst & Src
        ld              [hl+], a        ; store out LOW byte of Dst & Src

        ld              a, d            ; a <- HIGH byte of Src

        jr              z, .ChkZero

        and             [hl]            ; a = HIGH byte of Dst & Src
        ld              [hl], a         ; store out HIGH byte of Dst & Src
.ChkSign
        res             2, b            ; reset Zero (bit #2)

        add             a, a            ; CY = bit #7 (sign bit)
        jr              nc, .NoSign

        set             3, b            ; set Sign (bit #3)

        RETURN          \1              ; return, enabling interrupts

.ChkZero

        and             [hl]            ; a = HIGH byte of Dst & Src
        ld              [hl], a         ; store out HIGH byte of Dst & Src

        jr              nz, .ChkSign

        set             2, b            ; set Zero (bit #2)
.NoSign
        res             3, b            ; reset Sign (bit #3)

        RETURN          \1              ; return, enabling interrupts

        ENDM

; **
; AND_RD_8:    AND [RD] with A.
;
; Parameters:  a  = value to AND w/ [RD]
;              hl = (pointer to) RD
;
; Modifies:    a  = LOW byte of Result
;              hl = (pointer to) RD + 1
;

AND_RD_8: MACRO

        and             [hl]            ; a = LOW byte of Dst & Src
        ld              [hl+], a        ; store out LOW byte of Dst & Src
        ld              [hl], 0         ; store out HIGH byte of Dst & Src

        RETURN          \1              ; return, enabling interrupts

        ENDM

; **
; AND_RD_8_F:  AND [RD] with a, setting the Sign (always reset) and Zero
;              flags accordingly.
;
; Parameters:  a  = 8-bit value to AND w/ [RD]
;              hl = (pointer to) RD
;

AND_RD_8_F: MACRO

        res             3, b            ; reset Sign (bit #3)

        and             [hl]            ; a = LOW byte of Dst & Src
        ld              [hl+], a        ; store out LOW byte of Dst & Src

        jr              z, .Zero

        ld              [hl], 0         ; store out HIGH byte of Dst & Src

        res             2, b            ; reset Zero (bit #2)

        RETURN          \1              ; return, enabling interrupts

.Zero

        ld              [hl], a         ; store out HIGH byte of Dst & Src

        set             2, b            ; set Zero (bit #2)

        RETURN          \1              ; return, enabling interrupts

        ENDM

; **
; XOR_RD:      XOR [RD] with DA.
;
; Parameters:  da = (16-bit) value to be XORed w/ [RD]
;              hl = (pointer to) RD
;

XOR_RD: MACRO

        xor             [hl]            ; a = LOW byte of [RD] ^ Src
        ld              [hl+], a        ; store out LOW byte of [RD] ^ Src

        ld              a, d            ; a <- HIGH byte of Src

        xor             [hl]            ; a = HIGH byte of [RD] ^ Src
        ld              [hl], a         ; store out HIGH byte of [RD] ^ Src

        RETURN          \1              ; return, enabling interrupts

        ENDM

; **
; XOR_RD_F:    XOR [RD] with DA, setting the Sign and Zero flags accordingly.
;
; Parameters:  da = the value to XOR w/ [RD]
;              hl = (pointer to) RD
;

XOR_RD_F: MACRO

        xor             [hl]            ; a = LOW byte of [RD] ^ Src
        ld              [hl+], a        ; store out LOW byte of [RD] ^ Src

        ld              a, d            ; a <- HIGH byte of Src

        jr              nz, .NoZero\@

        xor             [hl]            ; a = HIGH byte of [RD] ^ Src
        ld              [hl], a         ; store out HIGH byte of [RD] ^ Src

        jr              nz, .ChkSign\@

        set             2, b            ; set Zero (bit #2)
.NoSign\@
        res             3, b            ; reset Sign (bit #3)

        RETURN          \1              ; return, enabling interrupts

.NoZero\@
        xor             [hl]            ; a = HIGH byte of [RD] ^ Src
        ld              [hl], a         ; store out HIGH byte of [RD] ^ Src
.ChkSign\@
        res             2, b            ; reset Zero (bit #2)

        add             a, a            ; CY = bit #7 (sign bit)
        jr              nc, .NoSign\@

        set             3, b            ; set Sign (bit #3)

        RETURN          \1              ; return, enabling interrupts

        ; ^ 24 bytes

        ENDM

; **
; XOR_RD_8:    XOR [RD] with A.
;
; Parameters:  a  = (8-bit) value to XOR w/ [RD]
;              hl = (pointer to) RD
;
; Modifies:    a  = LOW byte of Result
;

XOR_RD_8: MACRO

        xor             [hl]            ; a = LOW byte of Dst ^ Src
        ld              [hl], a         ; store out LOW byte of Dst ^ Src

        RETURN          \1              ; return, enabling interrupts

        ; ^ 8 cycles, 3 bytes

        ENDM

; **
; XOR_RD_8_F:  XOR [RD] with A, setting the Sign and Zero flags accordingly.
;
; Parameters:  a  = (8-bit) value to XOR w/ [RD]
;              hl = (pointer to) RD
;
; Modifies:    a, hl
;

XOR_RD_8_F: MACRO

        xor             [hl]            ; a = LOW byte of Dst ^ Src
        ld              [hl+], a        ; store out LOW byte of Dst ^ Src

        jr              z, .ChkZero\@

        ld              a, [hl]         ; a <- HIGH byte of Dst
.ChkSign\@
        res             2, b            ; reset Zero (bit #2)

        add             a, a            ; is sign (bit #7) set ?
        jr              nc, .NoSign\@

        set             3, b            ; set Sign (bit #3)

        RETURN          \1              ; return, enabling interrupts

.ChkZero\@

        or              [hl]            ; is HIGH byte of Dst = 0 too ?
        jr              nz, .ChkSign\@

        set             2, b            ; set Zero (bit #2)
.NoSign\@
        res             3, b            ; reset Sign (bit #3)

        RETURN          \1              ; return, enabling interrupts

        ENDM

; **
; DOREG:       Load DA w/ [RS], and HL w/ RD.
;
; Parameters:  l = LOW byte of RS
;              h = LOW byte of RD
;
; Modifies:    da = [RS]
;              e  = l
;              hl = (pointer to) RD
;

DOREG:  MACRO

        IF              \1 & mpDI != 0
        di                              ; don't allow VBlankFunc to interrupt
        ENDC

        ld              e, h            ; save LOW byte of RD in e
        ld              h, c            ; hl -> RS

        ld              a, [hl+]        ; a <- LOW byte of [RS]
        ld              d, [hl]         ; d <- HIGH byte of [RS]

        ld              l, e            ; hl -> RD

        ; ^ 7/8 cycles

        ENDM

; **
; DOREG_DR7:   Load DA w/ [RS], and load HL w/ [R7] (real).
;
; Parameters:  l  = LOW byte of RS
;
; Modifies:    hl = [R7] (real)
;              da = [RS]
;

DOREG_DR7: MACRO

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RS

        ld              a, [hl+]        ; a <- LOW byte of [RS]
        ld              d, [hl]         ; d <- HIGH byte of [RS]

        ldhl            sp, 0           ; hl = [R7] (real)

        ENDM

; **
; DOREG_SR7:   Load DA w/ [R7] (real) >> 1, and load HL w/ RD.
;
; Parameters:  e  = LOW byte of RD
;
; Modifies:    da = [R7] (real) >> 1
;              hl = (pointer to) RD
;

DOREG_SR7: MACRO

        di                              ; don't allow VBlankFunc to interrupt

        ldhl            sp, 0

        ld              d, h
        ld              a, l            ; da = [R7] (real)

        srl             d
        rra                             ; da = [R7] (real) >> 1

        ld              l, e
        ld              h, c            ; hl -> RD (in Echo RAM)

        ENDM

; **
; DODIRECT:    Disable interrupts, load DA w/ [ADDR], and HL w/ RD.
;
; Parameters:  e  = LOW byte of RD
;
; Modifies:    da = [ADDR]
;              hl = (pointer to) RD
;

DODIRECT: MACRO

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- ADDR (real)

        ld              a, [hl+]
        ld              d, [hl]         ; da <- [ADDR]

        ld              l, e
        ld              h, c            ; hl -> RD

        ENDM

; **
; DODIRECT_R7:  Disable interrupts, load HL w/ ADDR, and DA with
;               [R7] (real) >> 1.
;
; Modifies:     da = [R7] (real) >> 1
;               hl = (pointer to) ADDR
;

DODIRECT_R7: MACRO

        ldhl            sp, 2           ; hl = [R7] (real)

        di                              ; don't allow VBlankFunc to interrupt

        ld              d, h
        ld              a, l            ; da = hl

        pop             hl              ; hl <- ADDR (real)

        srl             d
        rra                             ; da = [R7] (real) >> 1

        ENDM

; **
; DOIMPLIED_R13:  Load DA w/ [[RM]] taking into account the SDBD state, and
;                 load HL w/ RD.
;
; Parameters:     l  = LOW byte of RM + 1
;                 h  = LOW byte of RD
;
; Modifies:       a  = LOW byte of [[RM]]
;                 d  = LOW  byte of [[RM]] if SDBD set
;                      HIGH byte of [[RM]] if SDBD not set
;                 e  = l
;                 hl = (pointer to) RD
;

DOIMPLIED_R13: MACRO

        di                              ; don't allow VBlankFunc to interrupt

        ld              e, h            ; save LOW byte of RD in e
        ld              h, c            ; hl -> RM + 1

        IF              R13_SDBD == 1

        ld              a, [SDBD]
        sub             c               ; is SDBD = cROMB0 & $FF ($D3) ?
        jr              nz, .NoSDBD     ; no, so not in SDBD mode

        ld              [SDBD], a       ; clear SDBD

        ld              a, [hl-]        ; a <- HIGH byte of [RM]
        ld              d, [hl]         ; d <- LOW byte of [RM]

        PEEK            mpSDBD_R13      ; da <- [[RM]]

        ld              d, a            ; SDBD mode, HIGH byte = LOW byte

        jr              .SDBD

        ENDC

.NoSDBD

        ld              a, [hl-]        ; a <- HIGH byte of [RM]
        ld              d, [hl]         ; d <- LOW byte of [RM]

        PEEK            0, d            ; da <- [[RM]]
.SDBD
        ld              l, e
        ld              h, c            ; hl -> RD

        ENDM

; **
; DOIMPLIED_R13_R7:  Load HL w/ [R7] (real) >> 1, and load DA w/ [[RM]].
;
; Parameters:        l  = LOW byte of RM + 1
;
; Modifies:          da = [[RM]]
;                    hl = [R7] (real) >> 1
;

DOIMPLIED_R13_R7: MACRO

        ; NOTE:  It is assumed that an SDBD would never occur with this
        ;        instruction, since the HIGH byte would merely receive a
        ;        copy of the LOW, which would unlikely be used as a value
        ;        to add to or subtract from R7.

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RM + 1

        ld              a, [hl-]        ; a <- HIGH byte of [RM]
        ld              d, [hl]         ; d <- LOW byte of [RM]

        PEEK            0, d            ; da <- [[RM]]

        ldhl            sp, 0           ; hl = [R7] (real)

        srl             h
        rr              l               ; hl = [R7] (real) >> 1

        ENDM

; **
; DOIMPLIED_R45:  Load DA w/ [[RM]] and increment [RM] (either R4 or R5)
;                 taking into account the SDBD state, and then load HL
;                 w/ RD.
;
; Parameters:     l  = LOW byte of RM
;                 h  = LOW byte of RD
;
; Modifies:       a  = LOW byte of [[RM]]
;                 e  = l
;                 d  = HIGH byte of [[RM]]      if SDBD is not set
;                      LOW  byte of [[RM] + 1]  if SDBD is set
;                 hl = (pointer to) RD
;

DOIMPLIED_R45: MACRO

        di                              ; don't allow VBlankFunc to interrupt

        ld              e, h            ; save LOW byte of RD in e
        ld              h, c            ; hl -> RM

        ld              a, [SDBD]
        sub             c               ; is SDBD = cROMB0 & $FF ($D3) ?
        jr              nz, .NoSDBD     ; no, so not in SDBD mode

        ld              [SDBD], a       ; clear SDBD

        ld              a, [hl]
        ld              d, a            ; d <- LOW byte of [RM]
        add             2
        ld              [hl+], a        ; store out LOW byte of [RM] + 2

        ld              a, [hl]         ; * a <- HIGH byte of [RM]
        jr              nc, .NoCarry\@

        inc             [hl]            ; increment HIGH byte of [RM]
.NoCarry\@
        PEEK            mpSDBD, d       ; da <- [[RM]]

        jr              .SDBD

.NoSDBD

        ld              a, [hl]
        ld              d, a            ; d <- LOW byte of [RM]
        inc             a
        ld              [hl+], a        ; store out LOW byte of [RM] + 1

        ld              a, [hl]         ; * a <- HIGH byte of [RM]
        jr              nz, .NoZero

        inc             [hl]            ; increment HIGH byte of [RM]
.NoZero
        PEEK            0, d            ; da <- [[RM]]
.SDBD
        ld              l, e
        ld              h, c            ; hl -> RD

        ENDM

; **
; DOIMPLIED_R45_R7:  Load HL w/ [R7] (real) >> 1, load DA w/ [[RM]],
;                    and increment [RM] (either R4 or R5) taking into
;                    account the SDBD state.
;
; Parameters:        l  = LOW byte of RM
;
; Modifies:          a  = LOW byte of [[RM]]
;                    d  = HIGH byte of [[RM]]      if SDBD is not set
;                         LOW  byte of [[RM] + 1]  if SDBD is set
;                    hl = [R7] (real) >> 1
;

DOIMPLIED_R45_R7: MACRO

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RM

        ld              a, [SDBD]
        sub             c               ; is SDBD = 1 ?
        jr              nz, .NoSDBD

        ld              [SDBD], a       ; clear SDBD

        ld              a, [hl]
        ld              d, a            ; d <- LOW byte of [RM]
        add             2
        ld              [hl+], a        ; store out LOW byte of [RM] + 2

        ld              a, [hl]         ; a <- HIGH byte of [RM]
        jr              nc, .NoCarry

        inc             [hl]            ; increment HIGH byte of [RM]
.NoCarry
        PEEK            mpSDBD, d       ; da <- [[RM]]

        jr              .SDBD_Cont

.NoSDBD

        ld              a, [hl]
        ld              d, a            ; d <- LOW byte of [RM]
        inc             a
        ld              [hl+], a        ; store out LOW byte of [RM] + 1

        ld              a, [hl]         ; a <- HIGH byte of [RM]
        jr              nz, .NoZero

        inc             [hl]            ; increment HIGH byte of [RM]
.NoZero
        PEEK            0, d            ; da <- [[RM]]
.SDBD_Cont
        ldhl            sp, 0           ; hl = [R7] (real)

        ENDM

; **
; DOIMPLIED_R6:  Disable interrupts, load DA w/ [[R6] - 1], decrement
;                [R6], and load HL w/ RD.
;
; Parameters:    e  = LOW byte of RD
;
; Modifies:      da = [[R6] - 1]
;                hl = (pointer to) RD
;

DOIMPLIED_R6: MACRO

        di                              ; don't allow VBlankFunc to interrupt

        ld              hl, R6          ; hl -> R6

        dec             [hl]            ; dec LOW byte of [R6]

        ld              l, [hl]         ; l <- LOW byte of [R6] - 1
        inc             l

        ld              a, [hl-]        ; a <- LOW byte of [[R6] - 1]
        inc             h
        ld              d, [hl]         ; d <- HIGH byte of [[R6] - 1]

        ld              l, e
        ld              h, c            ; hl -> RD

        ; ^ 17 cycles, 12 bytes (vs. 18 cycles, 13 bytes)

        ENDM

; **
; DOIMPLIED_R6_R7:  Load DA w/ [[R6] - 1], decrement [R6], and load HL w/
;                   [R7] (real) >> 1.
;
; Modifies:         da = [[R6] - 1]
;                   hl = [R7] (real) >> 1
;

DOIMPLIED_R6_R7: MACRO

        ld              hl, R6          ; hl -> R6

        di                              ; don't allow VBlankFunc to interrupt

        dec             [hl]            ; dec LOW byte of [R6]

        ld              l, [hl]         ; l <- LOW byte of [R6] - 1
        inc             l

        ld              a, [hl-]        ; a <- LOW byte of [[R6] - 1]
        inc             h
        ld              d, [hl]         ; d <- HIGH byte of [[R6] - 1]

        ldhl            sp, 0           ; hl = [R7] (real)

        ; ^ 18 cycles, 12 bytes (vs. 19 cycles, 13 bytes)

        ENDM

; **
; DOIMMEDIATE_16:  Load DA with XXYY (16-bit parameter), load HL with RD,
;                  and set BC = [R7] + 4.
;
; Parameters:      None.
;

DOIMMEDIATE_16: MACRO

        di                              ; don't allow VBlankFunc to interrupt

        pop             de              ; de <- AABB
        pop             hl              ; hl <- RD
        pop             af              ; af <- BB00

        ; ^ 10 cycles, 4 bytes

        ; OBJ_0: $AABB  AABB  = OpCode for 16-bit IMMEDIATE instruction
        ; OBJ_1: $XXQQ    QQ  = LOW byte of [R7] + 4
        ;                 XX  = HIGH byte of immediate 16-bit value
        ; OBJ_2: $RDYY    YY  = LOW byte of immediate 16-bit value
        ;                 RD  = LOW byte of (pointer to) RD
        ; OBJ_3: $0000  0000  = unused

;        ld              l, c
;        ld              h, b            ; hl -> HIGH byte of OBJ_0
;
;        inc             l               ; pos @ LOW byte of OBJ_1
;
;        ld              c, [hl]         ; c <- LOW byte of OBJ_1 (QQ)
;        inc             l
;
;        ld              d, [hl]         ; d <- HIGH byte of OBJ_1 (XX)
;        inc             l
;
;        ld              a, [hl+]        ; a <- LOW byte of OBJ_2 (YY)
;
;        ld              l, [hl]         ; l <- LOW byte of RD
;        ld              h, R0 >> 8

        ; ^ 15 cycles, 11 bytes

        ENDM

; **
; DOIMMEDIATE_10:  Load DE with AABB (10-bit parameter), and load HL w/ RD.
;
; Parameters:      l  = LOW byte of RD
;
; Modifies:        de = AABB (10-bit parameter)
;                  hl = (pointer to) RD
;
; Costs:           5 cycles, 3 bytes.
;

DOIMMEDIATE_10: MACRO

        di                              ; don't allow VBlankFunc to interrupt

        pop             de              ; de <- OBJ_01 (AABB)

        ld              h, c            ; hl -> RD

        ENDM

; **
; DOIMMEDIATE_8:  Load H with BB (8-bit parameter), load L with RD.
;
; Modifies:       l = LOW byte of (pointer to) RD
;                 h = 8-bit value
;

DOIMMEDIATE_8: MACRO

        ; OBJ_00: $AABB  $AABB = OpCode for 8-bit IMMEDIATE instruction
        ; OBJ_01: $YYRD  RD    = LOW byte of (pointer to) RD
        ;                YY    = 8-bit value

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; l <- LOW byte of OBJ_1 (RD)
                                        ; h <- HIGH byte of OBJ_2 (YY)

        ENDM
