; **
; Intellivision for GameBoy Color v1.1
; Copyright(C) 2000,03 By Carl Mueller, Jr.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

; ---------------------------------------------------------------------------
;  .::** Miscellaneous routines that don't need to be in HOME section **::.
; ---------------------------------------------------------------------------

; **
; STIC_Analyze:  Sets up the LocalOAM of sprites belonging to moving objects
;                if they have their "Re-Analyze" bit set.  It writes the
;                appropriate X and Y coordinates and attributes for each of
;                the MO's 4 sprites, and then adds the MO's address to
;                MOVELIST if its ReCopy bit is set.
;
; Parameters:    [rSVBK] must be = 5 (#/WRAM bank that contains LocalOAM
;                and MOVELIST).
;

STIC_Analyze:

         ; first, update the LocalOAM of sprites belonging to MO's that have
         ; their "Re-Analyze" bit set

         ld             hl, MO
         ld             de, MO_RECSIZE
.Analyze_Loop
         ld             a, [hl]         ; a <- MO attribute byte #0
         add            a, a            ; does MO need to be re-analyzed ?
         jr             c, .ReAnalyze

         add            hl, de          ; advance to next MO record
         jr             nz, .Analyze_Loop

         inc            a                   ; a = 1
         ld             [STIC_Analyzed], a  ; flag that LocalOAM is updated

         ; now queue MO's that have their ReCopy bits set in MOVELIST

         ld             l, MO & $FF + 3 ; hl -> MO[ MO # 0 + 3 ]
.ReCopy_Loop
         ld             a, [hl]         ; a <- MO attribute byte #1
         add            a, a            ; does MO need to be recopied ?
         jr             c, .ReCopy
.ReCopy_Cont
         add            hl, de
         jr             nz, .ReCopy_Loop

         ; find the 1st MO that isn't using all 3 colors of its palette,
         ; and set the "column" sprite's palette index to # of that MO

         ld             l, MO & $FF     ; hl -> MO[ MO # 0 ]
         ld             e, MO_RECSIZE - 3
.Search_MO
         ld             a, [hl+]        ; a <- MO attribute byte #0
         and            $02             ; is MO visable ?
         jr             z, .Available   ; no, so 3 out of 3 colors available

         inc            l               ; skip X-Coord.
         inc            l               ; skip Y-Coord.

         ld             a, [hl]         ; a <- MO attribute byte #1
         and            $03             ; are both YFULL and YSIZE clear ?
         jr             nz, .Available  ; no, so 2 out of 3 colors available

         add            hl, de          ; position at next MO
         jr             .Search_MO

.Available

         ld             a, l            ; a = MO # * 8 + 2 (or + 0)
         and            $38             ; isolate MO # * 8
         or             $84             ; set auto-advance, color #2
         ldh            [BC_OCPS], a

         rrca                           ; a = .1.mmm1.
         rrca                           ;   = ..1.mmm1
         rrca                           ;   = 1..1.mmm
         and            $07             ; * set (Y Flip & Pal. Bank - harmless)

         ld             hl, LocalOAM + 3  ; pos @ at sprite #0's OAM
         ld             e, 4              ; ea. OAM record is 4 bytes
         ld             c, 6              ; do 6 sprites
.SetPalette
         ld             [hl], a         ; set new Palette Index
         add            hl, de          ; position at next sprite's OAM
         dec            c
         jr             nz, .SetPalette

         ret

.ReCopy
         res            7, [hl]         ; reset ReCopy

         add            a, a            ; is MOVELIST-Queue bit already set ?
         jr             c, .ReCopy_Cont ; yes, so don't queue it again

         set            6, [hl]         ; set MOVELIST-Queue bit

         ld             a, [ML_BOTTOM]
         ld             c, a
         ld             b, MOVELIST >> 8 ; bc -> MOVELIST[ ML_BOTTOM ]

         di
         inc            a
         ld             [ML_BOTTOM], a  ; inc [ML_BOTTOM]
         ld             a, l            ; a = LOW of MO @ attribute byte #1
         ld             [bc], a         ; add it to MOVELIST
         ei

         jr             .ReCopy_Cont

.ReAnalyze

         res            7, [hl]         ; reset ReAnalyze

         ; MO ATTRIBUTE BYTE #0
         ;
         ;   07 06 05 04 03 02 01 00
         ;    a  0  1  p ym xm  v  c

         ld             d, l
         sla            d               ; d = MO # * 16

         ld             a, [hl+]        ; a <- MO attribute byte #0
         bit            1, a            ; is visability set ?
         jr             z, .Invisable   ; no, so just hide every sprite of MO

         ld             e, a            ; save MO attribute byte #0 in e

         ld             a, [rSCX_New]
         ld             c, a            ; c <- Scroll X

         ld             a, [hl+]        ; a <- X-Coord.
         dec		a       	; include zero in comparison
         cp		168 - 1		; is X = 0, or in 168..255 ?
         jr		nc, .Invisable	; yes, so hide every sprite of MO
         inc		a		; restore X-Coord.
         sub		c
         ld             c, a            ; c = X-Coord. - Scroll X

         ld             a, [rSCY_New]
         ld             b, a            ; b <- Scroll Y
         ld             a, [hl+]        ; a <- Y-Coord.
         sub            b
         ld             b, a            ; b = Y-Coord. - Scroll Y

         ld             l, [hl]         ; l <- MO attribute byte #1

         ld             a, e            ; a = MO attribute byte #0
         rrca                           ; put P, YM, and XM in bits 1-3
         and            $0E             ; isolate only those bits
         ld             h, a            ; save in h for later
         or             d               ; put MO # in bits 4-6
         swap           a               ; P, YM, and XM in bits 5-7
                                        ; Palette Index in bits 0-2
         ld             e, a            ; save Attribute in e

         ; MO ATTRIBUTE BYTE #1
         ;
         ;   07 06 05 04 03 02 01 00
         ;    r  q  1  0 yr xs ys yf

         ld             a, l            ; a <- MO attribute byte #1
         add            a, a
         add            a, a
         add            a, a            ; put YR, XS, YS and YF in bits 3-6
         res            3, h            ; reset P in case it was set
         or             h               ; combine with YM and XM in bits 1-2

         add            a, .JumpTable & $FF
         ld             l, a
         ld             a, .JumpTable >> 8
         adc            a, 0
         ld             h, a            ; hl -> .JumpTable[ a ]

         ld             a, [hl+]
         ld             h, [hl]
         ld             l, a

         ld             a, d            ; a = MO # * 16
         add            a, 8 * 4        ; skip the first 8 sprites

         jp             [hl]            ; jump to appropriate routine based
                                        ; on YR, XS, YS, YF, YM, and XM

.Invisable

         ; object is invisable, so the only thing necessary is to get the
         ; MO's sprites off the screen

         ld             h, LocalOAM >> 8
         ld             a, d            ; a = MO # * 16
         add            a, 8 * 4        ; skip the first 8 sprites
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             de, 4

         ld             [hl], d         ; zero Sprite #0's Y-Coord. (hide)
         add            hl, de
.Hide_3
         ld             [hl], d         ; zero Sprite #1's Y-Coord. (hide)
         add            hl, de
.Hide_2
         ld             [hl], d         ; zero Sprite #2's Y-Coord. (hide)
         add            hl, de
.Hide_1
         ld             [hl], d         ; zero Sprite #3's Y-Coord. (hide)
         add            hl, de
.Hide_0
         ld             a, l
         sub            8 * 4           ; a = MO # * 16
         rrca                           ;   = MO # * 8
         ld             l, a
         ld             h, MO >> 8      ; hl -> MO[ ( MO # + 1 ) * 8 ]

         ld             e, MO_RECSIZE

         jp             .Analyze_Loop   ; examine remaining MO records

.JumpTable

         DW             .Spr0_P8,    .Spr0_P8,     .Spr0_M4,     .Spr0_M4
         DW             .Spr0_P8,    .Spr0_P8,     .Spr0_NO,     .Spr0_NO
         DW             .Spr0_P8,    .Spr0_P8,     .Spr0_P8,     .Spr0_P8
         DW             .Spr01_P8,   .Spr01_P8,    .Spr01_P24,   .Spr01_P24
         DW             .Spr02_P8,   .Spr02_P8S,   .Spr02_M4,    .Spr02_M4S
         DW             .Spr02_P8,   .Spr02_P8S,   .Spr02_NO,    .Spr02_NOS
         DW             .Spr02_P8,   .Spr02_P8S,   .Spr02_P8,    .Spr02_P8S
         DW             .Spr0123_P8, .Spr0123_P8S, .Spr0123_P24, .Spr0123_P24S
         DW             .Spr0_P8,    .Spr0_P8,     .Spr0_NO,     .Spr0_NO
         DW             .Spr0_P8,    .Spr0_P8,     .Spr0_P8,     .Spr0_P8
         DW             .Spr01_P8,   .Spr01_P8,    .Spr01_P24,   .Spr01_P24
         DW             .Spr01_P8,   .Spr01_P8,    .Spr01_P56,   .Spr01_P56
         DW             .Spr02_P8,   .Spr02_P8S,   .Spr02_NO,    .Spr02_NOS
         DW             .Spr02_P8,   .Spr02_P8S,   .Spr02_P8,    .Spr02_P8S
         DW             .Spr0123_P8, .Spr0123_P8S, .Spr0123_P24, .Spr0123_P24S
         DW             .Spr0123_P8, .Spr0123_P8S, .Spr0123_P56, .Spr0123_P56S

.Spr0_P8

         ; Works for the following combinations:
         ;
         ; YR XS YS YF YM XM  V
         ;
         ;  0  0  0  0  0  0  1
         ;  0  0  0  0  0  1  1
         ;  0  0  0  1  0  0  1
         ;  0  0  0  1  0  1  1
         ;  0  0  1  0  0  0  1
         ;  0  0  1  0  0  1  1
         ;  0  0  1  0  1  0  1
         ;  0  0  1  0  1  1  1
         ;  1  0  0  0  0  0  1
         ;  1  0  0  0  0  1  1
         ;  1  0  0  1  0  0  1
         ;  1  0  0  1  0  1  1
         ;  1  0  0  1  1  0  1
         ;  1  0  0  1  1  1  1

         ld             h, LocalOAM >> 8
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             a, b            ; a <- MO's Y-Coord.
         add            8
         ld             [hl+], a        ; store out Sprite #0's Y-Coord.
         ld             [hl], c         ; store out Sprite #0's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #0's Attribute
         inc            l

         ld             de, 4

         jp             .Hide_3         ; hide Sprite #1, #2, and #3

.Spr0_M4

         ; Works for the following combinations:
         ;
         ; YR XS YS YF YM XM  V
         ;
         ;  0  0  0  0  1  0  1
         ;  0  0  0  0  1  1  1

         ld             h, LocalOAM >> 8
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             a, b            ; a <- MO's Y-Coord.
         sub            4
         ld             [hl+], a        ; store out Sprite #0's Y-Coord.
         ld             [hl], c         ; store out Sprite #0's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #0's Attribute
         inc            l

         ld             de, 4

         jp             .Hide_3         ; hide Sprite #1, #2, and #3

.Spr0_NO

         ; Works for the following combinations:
         ;
         ; YR XS YS YF YM XM  V
         ;
         ;  0  0  0  1  1  0  1
         ;  0  0  0  1  1  1  1
         ;  1  0  0  0  1  0  1
         ;  1  0  0  0  1  1  1

         ld             h, LocalOAM >> 8
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             [hl], b         ; store out Sprite #0's Y-Coord.
         inc            l
         ld             [hl], c         ; store out Sprite #0's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #0's Attribute
         inc            l

         ld             de, 4

         jp             .Hide_3         ; hide Sprite #1, #2, and #3

.Spr01_P8

         ; Works for the following combinations:
         ;
         ; YR XS YS YF YM XM  V
         ;
         ;  0  0  1  1  0  0  1
         ;  0  0  1  1  0  1  1
         ;  1  0  1  0  0  0  1
         ;  1  0  1  0  0  1  1
         ;  1  0  1  1  0  0  1 (with only the top half displayed)
         ;  1  0  1  1  0  1  1 (with only the top half displayed)

         ld             h, LocalOAM >> 8
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             a, b            ; a <- MO's Y-Coord.
         add            8
         ld             [hl+], a        ; store out Sprite #0's Y-Coord.
         ld             [hl], c         ; store out Sprite #0's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #0's Attribute
         inc            l

         add            16
         ld             [hl+], a        ; store out Sprite #1's Y-Coord.
         ld             [hl], c         ; store out Sprite #1's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #1's Attribute
         inc            l

         ld             de, 4

         jp             .Hide_2         ; hide Sprite #2 and #3

.Spr01_P24

         ; Works for the following combinations:
         ;
         ; YR XS YS YF YM XM  V
         ;
         ;  0  0  1  1  1  0  1
         ;  0  0  1  1  1  1  1
         ;  1  0  1  0  1  0  1
         ;  1  0  1  0  1  1  1

         ld             h, LocalOAM >> 8
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             a, b            ; a <- MO's Y-Coord.
         add            24
         ld             [hl+], a        ; store out Sprite #0's Y-Coord.
         ld             [hl], c         ; store out Sprite #0's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #0's Attribute
         inc            l

         sub            16
         ld             [hl+], a        ; store out Sprite #1's Y-Coord.
         ld             [hl], c         ; store out Sprite #1's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #1's Attribute
         inc            l

         ld             de, 4

         jp             .Hide_2         ; hide Sprite #2 and #3

.Spr01_P56

         ; Works for the following combinations:
         ;
         ; YR XS YS YF YM XM  V
         ;
         ;  1  0  1  1  1  0  1 (with only the top half displayed)
         ;  1  0  1  1  1  1  1 (with only the top half displayed)

         ld             h, LocalOAM >> 8
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             a, b            ; a <- MO's Y-Coord.
         add            56
         ld             [hl+], a        ; store out Sprite #0's Y-Coord.
         ld             [hl], c         ; store out Sprite #0's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #0's Attribute
         inc            l

         sub            16
         ld             [hl+], a        ; store out Sprite #1's Y-Coord.
         ld             [hl], c         ; store out Sprite #1's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #1's Attribute
         inc            l

         ld             de, 4

         jp             .Hide_2         ; hide Sprite #2 and #3

.Spr02_P8

         ; Works for the following combinations:
         ;
         ; YR XS YS YF YM XM  V
         ;
         ;  0  1  0  0  0  0  1
         ;  0  1  0  1  0  0  1
         ;  0  1  1  0  0  0  1
         ;  0  1  1  0  1  0  1
         ;  1  1  0  0  0  0  1
         ;  1  1  0  1  0  0  1
         ;  1  1  0  1  1  0  1

         ld             h, LocalOAM >> 8
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             a, b            ; a <- MO's Y-Coord.
         add            8
         ld             [hl+], a        ; store out Sprite #0's Y-Coord.
         ld             [hl], c         ; store out Sprite #0's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #0's Attribute
         inc            l

         ld             d, 0

         ld             [hl], d         ; zero Sprite #1's Y-Coord. (hide)
         inc            l
         inc            l               ; skip X Coord.
         inc            l               ; skip Tile Index
         inc            l               ; skip Attribute

         ld             [hl+], a        ; store out Sprite #2's Y-Coord.
         ld             a, c            ; a <- MO's X-Coord.
         add            8
         ld             [hl+], a        ; store out Sprite #2's X-Coord.
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #2's Attribute
         inc            l

         ld             e, 4

         jp             .Hide_1         ; hide Sprite #3

.Spr02_P8S

         ; Works for the following combinations:
         ;
         ; YR XS YS YF YM XM  V
         ;
         ;  0  1  0  0  0  1  1
         ;  0  1  0  1  0  1  1
         ;  0  1  1  0  0  1  1
         ;  0  1  1  0  1  1  1
         ;  1  1  0  0  0  1  1
         ;  1  1  0  1  0  1  1
         ;  1  1  0  1  1  1  1

         ld             h, LocalOAM >> 8
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             a, b            ; a <- MO's Y-Coord.
         add            8
         ld             b, a
         ld             [hl+], a        ; store out Sprite #0's Y-Coord.
         ld             a, c            ; a <- MO's X-Coord.
         add            8
         ld             [hl+], a        ; store out Sprite #0's X-Coord.
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #0's Attribute
         inc            l

         ld             d, 0

         ld             [hl], d         ; zero Sprite #1's Y-Coord. (hide)
         inc            l
         inc            l               ; skip X Coord.
         inc            l               ; skip Tile Index
         inc            l               ; skip Attribute

         ld             [hl], b         ; store out Sprite #2's Y-Coord.
         inc            l
         ld             [hl], c         ; store out Sprite #2's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #2's Attribute
         inc            l

         ld             e, 4

         jp             .Hide_1         ; hide Sprite #3

.Spr02_M4

         ; Works for the following combinations:
         ;
         ; YR XS YS YF YM XM  V
         ;
         ;  0  1  0  0  1  0  1

         ld             h, LocalOAM >> 8
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             a, b            ; a <- MO's Y-Coord.
         sub            4
         ld             [hl+], a        ; store out Sprite #0's Y-Coord.
         ld             [hl], c         ; store out Sprite #0's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #0's Attribute
         inc            l

         ld             d, 0

         ld             [hl], d         ; zero Sprite #1's Y-Coord. (hide)
         inc            l
         inc            l               ; skip X Coord.
         inc            l               ; skip Tile Index
         inc            l               ; skip Attribute

         ld             [hl+], a        ; store out Sprite #2's Y-Coord.
         ld             a, c            ; a <- MO's X-Coord.
         add            8
         ld             [hl+], a        ; store out Sprite #2's X-Coord.
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #2's Attribute
         inc            l

         ld             e, 4

         jp             .Hide_1         ; hide Sprite #3

.Spr02_M4S

         ; Works for the following combinations:
         ;
         ; YR XS YS YF YM XM  V
         ;
         ;  0  1  0  0  1  1  1

         ld             h, LocalOAM >> 8
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             a, b            ; a <- MO's Y-Coord.
         sub            4
         ld             b, a
         ld             [hl+], a        ; store out Sprite #0's Y-Coord.
         ld             a, c            ; a <- MO's X-Coord.
         add            8
         ld             [hl+], a        ; store out Sprite #0's X-Coord.
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #0's Attribute
         inc            l

         ld             d, 0

         ld             [hl], d         ; zero Sprite #1's Y-Coord. (hide)
         inc            l
         inc            l               ; skip X Coord.
         inc            l               ; skip Tile Index
         inc            l               ; skip Attribute

         ld             [hl], b         ; store out Sprite #2's Y-Coord.
         inc            l
         ld             [hl], c         ; store out Sprite #2's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #2's Attribute
         inc            l

         ld             e, 4

         jp             .Hide_1         ; hide Sprite #3

.Spr02_NO

         ; Works for the following combinations:
         ;
         ; YR XS YS YF YM XM  V
         ;
         ;  0  1  0  1  1  0  1
         ;  1  1  0  0  1  0  1

         ld             h, LocalOAM >> 8
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             [hl], b         ; store out Sprite #0's Y-Coord.
         inc            l
         ld             [hl], c         ; store out Sprite #0's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #0's Attribute
         inc            l

         ld             d, 0

         ld             [hl], d         ; zero Sprite #1's Y-Coord. (hide)
         inc            l
         inc            l               ; skip X Coord.
         inc            l               ; skip Tile Index
         inc            l               ; skip Attribute

         ld             [hl], b         ; store out Sprite #2's Y-Coord.
         inc            l
         ld             a, c            ; a <- MO's X-Coord.
         add            8
         ld             [hl+], a        ; store out Sprite #2's X-Coord.
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #2's Attribute
         inc            l

         ld             e, 4

         jp             .Hide_1         ; hide Sprite #3

.Spr02_NOS

         ; Works for the following combinations:
         ;
         ; YR XS YS YF YM XM  V
         ;
         ;  0  1  0  1  1  1  1
         ;  1  1  0  0  1  1  1

         ld             h, LocalOAM >> 8
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             [hl], b         ; store out Sprite #0's Y-Coord.
         inc            l
         ld             a, c            ; a <- MO's X-Coord.
         add            8
         ld             [hl+], a        ; store out Sprite #0's X-Coord.
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #0's Attribute
         inc            l

         ld             d, 0

         ld             [hl], d         ; zero Sprite #1's Y-Coord. (hide)
         inc            l
         inc            l               ; skip X Coord.
         inc            l               ; skip Tile Index
         inc            l               ; skip Attribute

         ld             [hl], b         ; store out Sprite #2's Y-Coord.
         inc            l
         ld             [hl], c         ; store out Sprite #2's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #2's Attribute
         inc            l

         ld             e, 4

         jp             .Hide_1         ; hide Sprite #3

.Spr0123_P8

         ; Works for the following combinations:
         ;
         ; YR XS YS YF YM XM  V
         ;
         ;  0  1  1  1  0  0  1
         ;  1  1  1  0  0  0  1
         ;  1  1  1  1  0  0  1 (with only top half displayed)

         ld             h, LocalOAM >> 8
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             a, b            ; a <- MO's Y-Coord.
         add            8
         ld             [hl+], a        ; store out Sprite #0's Y-Coord.
         ld             [hl], c         ; store out Sprite #0's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #0's Attribute
         inc            l

         add            16
         ld             b, a
         ld             [hl+], a        ; store out Sprite #1's Y-Coord.
         ld             [hl], c         ; store out Sprite #1's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #1's Attribute
         inc            l

         sub            16
         ld             [hl+], a        ; store out Sprite #2's Y-Coord.
         ld             a, c            ; a <- MO's X-Coord.
         add            8
         ld             [hl+], a        ; store out Sprite #2's X-Coord.
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #2's Attribute
         inc            l

         ld             [hl], b         ; store out Sprite #3's Y-Coord.
         inc            l
         ld             [hl+], a        ; store out Sprite #2's X-Coord.
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #2's Attribute
         inc            l

         ld             d, 0            ; *

         jp             .Hide_0

.Spr0123_P8S

         ; Works for the following combinations:
         ;
         ; YR XS YS YF YM XM  V
         ;
         ;  0  1  1  1  0  1  1
         ;  1  1  1  0  0  1  1
         ;  1  1  1  1  0  1  1 (with only top half displayed)

         ld             h, LocalOAM >> 8
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             a, b            ; a <- MO's Y-Coord.
         add            8
         ld             [hl+], a        ; store out Sprite #0's Y-Coord.
         ld             a, c            ; a <- MO's X-Coord.
         add            8
         ld             [hl+], a        ; store out Sprite #0's X-Coord.
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #0's Attribute
         inc            l

         ld             a, b            ; a <- MO's Y-Coord.
         add            24
         ld             [hl+], a        ; store out Sprite #1's Y-Coord.
         ld             a, c            ; a <- MO's X-Coord.
         add            8
         ld             [hl+], a        ; store out Sprite #1's X-Coord.
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #1's Attribute
         inc            l

         ld             a, b
         add            8
         ld             [hl+], a        ; store out Sprite #2's Y-Coord.
         ld             [hl], c         ; store out Sprite #2's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #2's Attribute
         inc            l

         add            16
         ld             [hl+], a        ; store out Sprite #3's Y-Coord.
         ld             [hl], c         ; store out Sprite #2's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #2's Attribute
         inc            l

         ld             d, 0            ; *

         jp             .Hide_0

.Spr0123_P24

         ; Works for the following combinations:
         ;
         ; YR XS YS YF YM XM  V
         ;
         ;  0  1  1  1  1  0  1
         ;  1  1  1  0  1  0  1

         ld             h, LocalOAM >> 8
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             a, b            ; a <- MO's Y-Coord.
         add            24
         ld             [hl+], a        ; store out Sprite #0's Y-Coord.
         ld             [hl], c         ; store out Sprite #0's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #0's Attribute
         inc            l

         sub            16
         ld             b, a
         ld             [hl+], a        ; store out Sprite #1's Y-Coord.
         ld             [hl], c         ; store out Sprite #1's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #1's Attribute
         inc            l

         add            16
         ld             [hl+], a        ; store out Sprite #2's Y-Coord.
         ld             a, c            ; a <- MO's X-Coord.
         add            8
         ld             [hl+], a        ; store out Sprite #2's X-Coord.
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #2's Attribute
         inc            l

         ld             [hl], b         ; store out Sprite #3's Y-Coord.
         inc            l
         ld             [hl+], a        ; store out Sprite #2's X-Coord.
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #2's Attribute
         inc            l

         ld             d, 0            ; *

         jp             .Hide_0

.Spr0123_P24S

         ; Works for the following combinations:
         ;
         ; YR XS YS YF YM XM  V
         ;
         ;  0  1  1  1  1  1  1
         ;  1  1  1  0  1  1  1

         ld             h, LocalOAM >> 8
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             a, b            ; a <- MO's Y-Coord.
         add            24
         ld             [hl+], a        ; store out Sprite #0's Y-Coord.
         ld             a, c            ; a <- MO's X-Coord.
         add            8
         ld             [hl+], a        ; store out Sprite #0's X-Coord.
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #0's Attribute
         inc            l

         ld             a, b
         add            8
         ld             [hl+], a        ; store out Sprite #1's Y-Coord.
         ld             a, c
         add            8
         ld             [hl+], a        ; store out Sprite #1's X-Coord.
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #1's Attribute
         inc            l

         ld             a, b
         add            24
         ld             [hl+], a        ; store out Sprite #2's Y-Coord.
         ld             [hl], c         ; store out Sprite #2's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #2's Attribute
         inc            l

         sub            16
         ld             [hl+], a        ; store out Sprite #3's Y-Coord.
         ld             [hl], c         ; store out Sprite #2's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #2's Attribute
         inc            l

         ld             d, 0            ; *

         jp             .Hide_0

.Spr0123_P56S

         ; Works for the following combinations:
         ;
         ; YR XS YS YF YM XM  V
         ;
         ;  1  1  1  1  1  1  1 (with only top half displayed)

         ld             h, LocalOAM >> 8
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             a, b            ; a <- MO's Y-Coord.
         add            56
         ld             [hl+], a        ; store out Sprite #0's Y-Coord.
         ld             a, c
         add            8
         ld             [hl+], a        ; store out Sprite #0's X-Coord.
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #0's Attribute
         inc            l

         ld             a, b            ; a <- MO's Y-Coord.
         add            40
         ld             [hl+], a        ; store out Sprite #1's Y-Coord.
         ld             a, c            ; a <- MO's X-Coord.
         add            8
         ld             [hl+], a        ; store out Sprite #1's X-Coord.
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #1's Attribute
         inc            l

         ld             a, b
         add            56
         ld             [hl+], a        ; store out Sprite #2's Y-Coord.
         inc            l
         ld             [hl], c         ; store out Sprite #2's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #2's Attribute
         inc            l

         sub            16
         ld             [hl+], a        ; store out Sprite #3's Y-Coord.
         ld             [hl], c         ; store out Sprite #2's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #2's Attribute
         inc            l

         ld             d, 0            ; *

         jp             .Hide_0

.Spr0123_P56

         ; Works for the following combinations:
         ;
         ; YR XS YS YF YM XM  V
         ;
         ;  1  1  1  1  1  0  1 (with only top half displayed)

         ld             h, LocalOAM >> 8
         ld             l, a            ; hl -> LocalOAM[ MO # * 16 ]

         ld             a, b            ; a <- MO's Y-Coord.
         add            56
         ld             b, a
         ld             [hl+], a        ; store out Sprite #0's Y-Coord.
         ld             [hl], c         ; store out Sprite #0's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #0's Attribute
         inc            l

         sub            16
         ld             [hl+], a        ; store out Sprite #1's Y-Coord.
         ld             [hl], c         ; store out Sprite #1's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #1's Attribute
         inc            l

         ld             [hl], b         ; store out Sprite #2's Y-Coord.
         inc            l
         ld             a, c            ; a <- MO's X-Coord.
         add            8
         ld             c, a
         ld             [hl+], a        ; store out Sprite #2's X-Coord.
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #2's Attribute
         inc            l

         ld             a, b            ; a <- MO's Y-Coord. - 56
         sub            16
         ld             [hl+], a        ; store out Sprite #3's Y-Coord.
         ld             [hl], c         ; store out Sprite #2's X-Coord.
         inc            l
         inc            l               ; skip Tile Index
         ld             [hl], e         ; store out Sprite #2's Attribute
         inc            l

         ld             d, 0            ; *

         jp             .Hide_0

; **
; GRAM_CheckFlags:  Check if any of the GRAM locations modified during
;                   the last call to Process will affect the background
;                   or moving objects by examining GRAM_Flags and take
;                   the appropriate measures.
;
; Parameters:       [rSVBK] must be = 6 (WRAM bank with MakeUp_GRAM,
;                   TILE, and COPYLIST).
;
; Modifies:         All registers.
;

GRAM_CheckFlags:

        IF              GRAM_MODIFIED == 1

        ld              bc, GRAM_Flags + GRAM_Flags >> 8

        ld              a, [bc]         ; a <- GRAM_Modified
        sub             b               ; is GRAM_Modified = $C5 ?
        ret             nz              ; no, so return

        ld              [bc], a         ; set GRAM_Modified = 0

        ELSE

        ld              b, GRAM_Flags >> 8

        ENDC

        ; 1st check if the modified GRAM locations will affect the MOs

        ld              hl, MO + 5
        ld              de, MO_RECSIZE + 2
.MO_Loop
        ld              a, [hl-]        ; a <- HIGH byte of MO's DataAddr
        sub             GRAM_8 >> 8     ; take off the GRAM base address
        jr              c, .MO_InGROM   ; if it was too much, must be in GROM

        add             [hl]            ; put LSBs of card index in bits 3-7
        rlca                            ; put MSBs of card index in bits 0-1
        ld              c, a            ; bc -> GRAM_Flags[ CardNdx ]

        dec             l               ; pos @ MO attribute byte #1

        ld              a, [bc]         ; a <- GRAM_Flags[ CardNdx ]
        cp              b               ; is GRAM_Flag "set" ?
        jr              z, .MO_ReCopy   ; yes, so MO needs to be recopied

        bit             3, [hl]         ; is YRES set ?
        jr              z, .MO_Next     ; no, so go on to the next MO

        ; YRES is set (using 2 cards), so check if _2nd_ card was written to

        add             $10             ; CY set if card #15, #31, #47, #63
        adc             a, d            ; = adc a, 0
        ld              c, a            ; bc -> GRAM_Flags[ CardNdx + 1 ]

        ld              a, [bc]         ; a <- GRAM_Flags[ CardNdx + 1 ]
        cp              b               ; is GRAM_Flag "set" ?
        jr              nz, .MO_Next    ; no, so don't recopy MO
.MO_ReCopy
        set             7, [hl]         ; set ReCopy (see STIC_Analyze)
.MO_Next
        add             hl, de          ; advance to next MO record

        ld              a, l
        sub             ( MO + MO_RECSIZE * 8 + 5 ) & $FF
        jr              nz, .MO_Loop

.Flags_Check

        ; find out which GRAM cards were modified by looping thru GRAM_Flags

        ld              c, d            ; hl -> GRAM_Flags @ 1st CardNdx
.Flags_Loop
        ld              a, [bc]         ; a <- "GRAM_Flag" for this card
        cp              b               ; is it = HIGH byte of GRAM_Flags ?
        jr              z, .Flag_Set    ; yes, so flag is considered set

        add             $10             ; CY set if card #15, #31, #47, #63
        adc             a, d            ; = adc a, 0
        ld              c, a            ; bc -> next card's "GRAM_Flag"

        cp              $04              ; is next CardNdx = 64 ?
        jr              nz, .Flags_Loop  ; no, so keep checking flags

        ret

.Flag_Set

        ; GRAM_Flag set, now setup a pointer to MakeUp_GRAM for current card

        ld              a, c            ; a = nnnn..NN

        ld              [bc], a         ; "clear" GRAM_Flag for this card

        and             $F0             ;   = nnnn....
        ld              l, a

        xor             c                 ; a = ......NN
        add             MakeUp_GRAM >> 8  ;   = 11.1..NN
        ld              h, a              ; hl -> MakeUp_GRAM[ CardNdx * 8 ]

        ; now loop through each possible makeup for card, checking for tiles

        ld              d, 16           ; iterate thru each possible makeup
.MakeUp_Loop
        ld              a, [hl+]        ; a <- TileNdx
        or              a               ; tile exist w/ current makeup ?
        jr              nz, .ChkInUseBy ; yes, so check if it's in use

        dec             d                 ; checked every possible combo ?
        jr              nz, .MakeUp_Loop  ; no, so keep checking

        ld              a, l            ; a = nnnn....
        sub             MakeUp_GRAM >> 8
        add             h               ;   = nnnn..NN

        ld              c, a
        ld              b, GRAM_Flags >> 8

        cp              $04              ; is next CardNdx = 64 ?
        jr              nz, .Flags_Loop  ; no, so keep checking flags

        ret

.MO_InGROM

        ; MO uses a GROM card (or cards), so its bitmap data wasn't modified

        dec             l               ; pos @ MO attribute byte #1
        add             hl, de          ; advance to next MO record

        ld              a, l
        sub             ( MO + MO_RECSIZE * 8 + 5 ) & $FF
        jr              nz, .MO_Loop

        jr              .Flags_Check    ; checked MOs, now check cards

.ChkInUseBy

        ; associated tile in VRAM found, but is it in currently being used ?

        ld              c, a
        ld              b, TileInUseBy >> 8  ; bc -> TileInUseBy[ TileNdx ]

        ld              a, [bc]         ; a <- #/cards using this tile
        or              a               ; is it <> 0 (at least 1 card) ?
        jr              nz, .ChkQueue   ; yes, so try to add it to COPYLIST

        dec             l               ; hl -> MakeUp_GRAM[ TileNdx ]
        ld              [hl+], a        ; "invalidate" tile in MakeUp table

        ;

        sla             c
        rla
        sla             c
        rla
        sla             c
        rla

        add             TILE >> 8
        ld              b, a            ; bc -> TILE[ TileNdx * 8 ]

        ld              a, MakeUp_DMY >> 8
        ld              [bc], a

        ; ^ 16 cycles

        dec             d                 ; checked every possible combo ?
        jr              nz, .MakeUp_Loop  ; no, so continue checking

        ld              a, l            ; a = nnnn....
        sub             MakeUp_GRAM >> 8
        add             h               ;   = nnnn..NN

        ld              c, a
        ld              b, GRAM_Flags >> 8

        cp              $04              ; is next CardNdx = 64 ?
        jr              nz, .Flags_Loop  ; no, so keep checking flags

        ret

.ChkQueue

        ; tile in VRAM found and in use, so try to add it to COPYLIST-Queue

        push            hl              ; * save MakeUpNdx

        ld              l, c
        ld              h, TILE >> 11   ; TILE aligned on multiples of 2048

        add             hl, hl
        add             hl, hl
        add             hl, hl          ; hl -> TILE[ TileNdx * 8 ]

        inc             l               ; skip HIGH byte of MakeUpNdx
        inc             l               ; skip LOW byte of MakeUpNdx

        bit             6, [hl]         ; is COPYLIST-Queue (bit #6) set ?
        jr              nz, .SkipQueue  ; yes, so don't queue it again

        set             6, [hl]         ; set COPYLIST-Queue (bit #6)

        ld              a, l
        ld              e, h            ; save TILE @ MakeUpBank + Queue
        di                              ; don't allow CopyToVRAM to interrupt
        ld              hl, CL_BOTTOM

        inc             [hl]            ; inc CL_BOTTOM

        ld              l, [hl]
        dec             l               ; l <- CL_BOTTOM (before increment)

        inc             h               ; hl -> COPYLIST[ CL_BOTTOM ]

        ld              [hl], a         ; store out LOW byte of TILE address
        inc             h
        ld              [hl], e         ; store out HIGH byte of TILE address
        ei                              ; allow CopyToVRAM to interrupt again

.SkipQueue

        pop             hl              ; restore MakeUpNdx

        dec             d                 ; checked every possible combo ?
        jr              nz, .MakeUp_Loop  ; no, so continue checking

        ld              a, l            ; a = nnnn....
        sub             MakeUp_GRAM >> 8
        add             h               ;   = nnnn..NN

        ld              c, a
        ld              b, GRAM_Flags >> 8

        cp              $04              ; is next CardNdx = 64 ?
        jp              nz, .Flags_Loop  ; no, so keep checking flags

        ret

; **
; MO_GetColors:  Converts the color _values_ stored in MO_FgCols and
;                MO_BgCols (the "background" color for the MO -- either
;                a fixed value or a dynamic one from the ColorStack) into
;                the appropriate RGB colors, and then stores them out to
;                MO_Colors.
;
; Parameters:    None.
;
; Modifies:      All registers.
;

MO_GetColors:

        ld              de, MO_FgCols   ; de -> MO_FgCols
        ld              b, 8            ; check colors for all 8 MOs
.Loop
        ld              a, [de]         ; a <- MO's FG color value (bits 4-7)
        rrca                            ; does RGB need to be re-looked up ?
        jr              c, .GetColor    ; yes, so handle it

        inc             e               ; pos @ next MO's FG color value

        dec             b               ; have we done all eight MOs ?
        jr              nz, .Loop       ; no, so keep looping

        ret

.GetColor

        add             a, a            ; reset ReAnalyze bit (CY = 1)
        ld              [de], a         ; and update the FG color byte

        ld              l, a            ; l = FG color << 4

        ld              a, e            ; a = 11...mmm

        rlca                            ;   = 1...mmm1
        ld              c, a            ; c = MO # * 2 + $81
        rlca                            ; a = ...mmm11       (CY = 1)
        adc             a, c            ;   = MO # * 6 + $85 (CY = 0)

        ld              c, a            ; c -> MO_Colors[ MO # * 6 + 5 ]

        set             3, e            ; pos @ MO_BgCols

        ld              a, [de]         ; a <- BG color in bits 0-3
        rra                             ; put LSB of BG color in CY
        rr              l               ; put LSB of BG color in bit #7
        ld              h, a            ; hl -> ObjPal[ FgCol * 8 + BgCol
                                        ;         * 128 ]

        ld              a, [hl+]        ; a <- HIGH byte of RGB color value
        ld              [c], a          ; store out to HIGH of MO_Colors #3
        dec             c

        ld              a, [hl+]        ; a <- LOW byte of RGB color value
        ld              [c], a          ; store out to LOW of MO_Colors #3
        dec             c

        dec             c
        dec             c               ; skip @ MO_Colors #2 (for now)

        ld              a, [hl+]        ; a <- HIGH byte of RGB color value
        ld              [c], a          ; store out to HIGH of MO_Colors #1
        dec             c

        ld              a, [hl+]        ; a <- LOW byte of RGB color value
        ld              [c], a          ; store out to LOW of MO_Colors #1
        inc             c

        inc             c               ; re-pos @ MO_Colors #2

        ld              a, [hl+]        ; a <- LOW byte of RGB color value
        ld              [c], a          ; store out to LOW of MO_Colors #2
        inc             c

        ld              a, [hl+]        ; a <- HIGH byte of RGB color value
        ld              [c], a          ; store out to HIGH of MO_Colors #2

        res             3, e            ; re-pos @ MO_FgCols

        inc             e               ; pos @ next MO's FG color value

        dec             b               ; have we done all eight MOs ?
        jr              nz, .Loop       ; no, so keep looping

        ret

