; **
; Intellivision for GameBoy Color v1.1
; Copyright(C) 2000,03 By Carl Mueller, Jr.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

        ; -------------------------------------------------------------------
        ;                   .::** TimerFunc **::.
        ; -------------------------------------------------------------------

TimerFunc_ROM::

        ld              [TF_M2 + 1], sp ; save SP at interrupt - 2

        ld              sp, tmfRegs + 4 * 2

        push            af
        push            bc
        push            de
        push            hl              ; save Regs in tmfRegs

        ld              sp, ayEnvVol

        ; -------------------------------------------------------------------
        ;          Set rAUDVOL and rAUDTERM based on ayEnvVol
        ; -------------------------------------------------------------------

        pop             bc              ; c <- ayEnvVol
                                        ; b <- ayEnvInc

        ld              l, c
        ld              h, ayVolToAudVol >> 8  ; hl -> ayVolToAudVol
                                               ;       [ ayEnvVol ]

        ld              a, [hl]         ; a <- output levels for SO1 and SO2

        ld              l, ayMask & $FF ; hl -> ayMask, followed by ayAudTerm

        cp              $0F             ; is AudVol = "-1" ?
        jr              z, .MuteSO2     ; yes, so SO2 needs to be "muted"

        ldh             [rAUDVOL], a    ; update SO1 and SO2 output levels

        ld              a, l            ; = ld a, $FF (don't mute SO2)
.MuteSO2
        ld              [hl-], a        ; keep SO2 muted in AY8914_Update
        and             [hl]            ; AND w/ last rAUDTERM value used

        ldh             [rAUDTERM], a   ; clear SO2 output enable bits

        ; if ayEnvInc = 0, ayEnvVol won't change any more, so stop the timer

        ld              a, b
        or              a               ; is ayEnvInc = 0 ?
        jr              z, .StopTimer   ; yes, so ayEnvVol won't change

        ; -------------------------------------------------------------------
        ;      Update ayEnvVol using ayEnvCtr, ayEnvPerCur and ayEnvInc
        ; -------------------------------------------------------------------

        pop             hl              ; hl <- ayEnvCtr

        ld              de, -436        ; de = 3579545 / 16 / 256 / 2
        add             hl, de
        jr              nc, .Env_CtrNeg ; if CY = 0, then ayEnvCtr was < 436
                                        ; if CY = 1, then ayEnvCtr was >= 436

        ld              a, l
        or              h               ; is ayEnvCtr = 0 exactly ?
        jr              z, .Env_CtrZero ; if ZF = 1, then ayEnvCtr was = 436

        ; ayEnvCtr positive, so just update ayEnvCtr (don't process "hits")

        push            hl              ; update ayEnvCtr

        jr              rTF_Exit

.Env_CtrNeg

        ; ayEnvCtr negative, so add ayEnvPeriod until it reaches or passes 0

        pop             de              ; de <- ayEnvPerCur

        ld              a, e
        or              d               ; is ayEnvPerCur = 0 ?

        ld              a, c            ; put ayEnvVol in a

        jr              z, .Env_PerZero ; yes, so process just one iteration

.Env_Loop

        add             b               ; add ayEnvVol, ayEnvInc
        cp              16              ; ayEnvVol >= 16 (end-of-envelope) ?
        ret             nc              ; if CY = 0, PC <- ayEnvEndHandler

        add             hl, de          ; add ayEnvCtr, ayEnvPerCur
        jr              nc, .Env_Loop   ; if CY = 0, ayEnvCtr still negative

        ; if ayEnvCtr _passed_ 0, we're done processing "hits"

        ld              c, a            ; save ayEnvVol in c (for updating)

        ld              a, l
        or              h               ; did ayEnvCtr hit 0 exactly ?
        jr              z, .Env_HitZero ; yes, so process one more iteration

        ld              sp, ayEnvCtr + 2

        push            hl              ; update envCtr
        push            bc              ; update envVol

        jr              rTF_Exit

.Env_CtrZero

        ; ayEnvCtr was = 436 and is now zero, so process just one iteration

        pop             de              ; de <- ayEnvPerCur

.Env_HitZero

        ; ayEnvCtr _reached_ 0 exactly, so process one more iteration

        ld              a, c            ; restore ayEnvVol from c

        add             b               ; add ayEnvVol, ayEnvInc
        cp              16              ; ayEnvVol >= 16 (end-of-envelope) ?
        ret             nc              ; if CY = 0, PC <- ayEnvEndHandler

        ld              sp, ayEnvCtr + 2

        push            de              ; update ayEnvCtr (=ayEnvPerCur)
        ldh             [ayEnvVol], a   ; update ayEnvVol

        jr              rTF_Exit

.Env_PerZero

        ; if ayEnvPerCur = 0, no reason to add ayEnvPerCur to ayEnvCtr

        add             b               ; add ayEnvVol, ayEnvInc
        cp              16              ; ayEnvVol >= 16 (end-of-envelope) ?
        ret             nc              ; if CY = 0, PC <- ayEnvEndHandler

        ld              sp, ayEnvCtr + 2

        push            hl              ; update envCtr
        ldh             [ayEnvVol], a   ; update envVol

        jr              rTF_Exit

.StopTimer

        ldh             [rTAC], a       ; ayEnvVol can't change, stop timer

        jr              rTF_Exit

        ; -------------------------------------------------------------------
        ;         End-of-envelope handlers (see ayEnvTypeToHandler)
        ; -------------------------------------------------------------------

rTF_EnvEnd_0:                           ; CONT: 0 ATAK: 0 ALTR: 0 HOLD: 0
rTF_EnvEnd_1:                           ; CONT: 0 ATAK: 0 ALTR: 0 HOLD: 1
rTF_EnvEnd_2:                           ; CONT: 0 ATAK: 0 ALTR: 1 HOLD: 0
rTF_EnvEnd_3:                           ; CONT: 0 ATAK: 0 ALTR: 1 HOLD: 1
rTF_EnvEnd_4:                           ; CONT: 0 ATAK: 1 ALTR: 0 HOLD: 0
rTF_EnvEnd_5:                           ; CONT: 0 ATAK: 1 ALTR: 0 HOLD: 1
rTF_EnvEnd_6:                           ; CONT: 0 ATAK: 1 ALTR: 1 HOLD: 0
rTF_EnvEnd_7:                           ; CONT: 0 ATAK: 1 ALTR: 1 HOLD: 1
rTF_EnvEnd_9:                           ; CONT: 1 ATAK: 0 ALTR: 0 HOLD: 1
rTF_EnvEnd_F:                           ; CONT: 1 ATAK: 1 ALTR: 1 HOLD: 1

        ; 0-3 = Decay only.  Sound drops from peak & stops.
        ; 4-7 = Attack only.  Sound rises to peak & stops.
        ; 9   = Initial decay, followed by sustain of zero volume. (=0-3)
        ; 15  = Initial attack, followed by sustain of zero volume. (=4-7)

        xor             a
        ldh             [ayEnvInc], a   ; set ayEnvInc = 0
        ldh             [ayEnvVol], a   ; set ayEnvVol = 0

        jr              rTF_Exit

rTF_EnvEnd_C:                           ; CONT: 1 ATAK: 1 ALTR: 0 HOLD: 0

        ; 12 = Attack only repeats indefinitely.

        xor             a
        ldh             [ayEnvVol], a   ; set ayEnvVol = 0

        ld              sp, ayEnvCtr + 2

        push            de              ; set ayEnvCtr = ayEnvPerCur

        jr              rTF_Exit

rTF_EnvEnd_D:                           ; CONT: 1 ATAK: 1 ALTR: 0 HOLD: 1

        ; 13 = Initial attack, followed by sustain of peak volume.

        sub             b
        ldh             [ayEnvVol], a   ; set ayEnvVol = 15
        xor             a
        ldh             [ayEnvInc], a   ; set ayEnvInc = 0

        jr              rTF_Exit

rTF_EnvEnd_A:                           ; CONT: 1 ATAK: 0 ALTR: 1 HOLD: 0
rTF_EnvEnd_E:                           ; CONT: 1 ATAK: 1 ALTR: 1 HOLD: 0

        ; 10 = Alternates indefinitely between attack and decay, starting
        ;      with decay.
        ; 14 = Alternates indefinitely between attack and decay, starting
        ;      with attack.

        sub             b               ; restore ayEnvVol (to 0 or 15)
        ld              c, a            ; and save it in c (for updating)

        ld              a, $FE
        xor             b               ; toggle ayEnvInc between -1 and 1
        ld              b, a            ; and save it in b (for updating)

        ld              sp, ayEnvCtr + 2

        push            de              ; set ayEnvCtr = ayEnvPerCur
        push            bc              ; update ayEnvInc and ayEnvVol

        jr              rTF_Exit

rTF_EnvEnd_B:                           ; CONT: 1 ATAK: 0 ALTR: 1 HOLD: 1

        ; 11 = Initial decay, followed by a sustain of peak volume.

        ld              a, $0F
        ldh             [ayEnvVol], a   ; set ayEnvVol = 15
        xor             a
        ldh             [ayEnvInc], a   ; set ayEnvInc = 0

        jr              rTF_Exit

rTF_EnvEnd_8:                           ; CONT: 1 ATAK: 0 ALTR: 0 HOLD: 0

        ; 8  = Decay only repeats indefinitely.

        ld              a, $0F
        ldh             [ayEnvVol], a   ; set ayEnvVol = 15

        ld              sp, ayEnvCtr + 2

        push            de              ; set ayEnvCtr = ayEnvPerCur

        ; -------------------------------------------------------------------
        ;                 Clean-up code for TimerFunc
        ; -------------------------------------------------------------------

rTF_Exit:

        ld              sp, tmfRegs

        pop             hl
        pop             de
        pop             bc
        pop             af              ; restore regs from tmfRegs

rTF_M2: ld              sp, 0           ; sp <- SP at interrupt - 2

        reti                            ; return, enabling interrupts

TimerFunc_ROM_End:

; **
; AY8914_Init:  Initialize everything having to do with AY8914 emulation
;               on the GameBoy, including necessary look-up tables in RAM
;               and initializing the GameBoy's sound registers.
;
; Parameters:   None.
;
; Modifies:     All registers.
;

AY8914_Init:

        ld              de, TimerFunc
        ld              hl, TimerFunc_ROM
        ld              bc, TimerFunc_ROM_End - TimerFunc_ROM

        call            MemCopy

        ; copy various tables to Low RAM, aligning them on multiples of 256

        ld              hl, ayVolToAudVol_ROM
        ld              de, ayVolToAudVol
        ld              bc, 16

        call            MemCopy

        ld              de, ayPerToAudPoly

.DupPerToAudPoly

        ld              hl, ayPerToAudPoly_ROM
        ld              c, 32

        call            MemCopy         ; setup ayPerToAudPoly from ROM

        ld              a, d                 ; a = HIGH byte of dst. address
        cp              ayPerToAudPoly >> 8  ; still on same page ?
        jr              z, .DupPerToAudPoly  ; yes, so fill it w/ duplicates

        ld              de, ayVolToAudEnv
        ld              c, 16

        call            MemCopy         ; setup ayVolToAudEnv from ROM

        ld              e, ayRegToHandler & $FF
        ld              c, 23 * 2

        call            MemCopy         ; setup ayRegToHandler from ROM

;.DupEnvTypeToHandler
;
;        ld              hl, ayEnvTypeToHandler_ROM
;        ld              c, 16 * 2
;
;        call            MemCopy         ; setup ayEnvTypeToHandler from ROM
;
;        ld              a, d                     ; a = HIGH of dst. address
;        cp              ayEnvTypeToHandler >> 8  ; still on same page ?
;        jr              z, .DupEnvTypeToHandler  ; yes, so fill w/ duplicates
;
;        ld              de, ayEnabledToAudTerm
;        ld              bc, 512
;
;        call            MemCopy         ; setup ayEnabledToAudTerm from ROM

        ; initialize sound registers and Wave Pattern RAM

        ld              de, rAUD1SWEEP
        ld              c, 23

        call            MemCopy         ; initialize the GB's sound registers

        xor             a
        ld              hl, _AUD3WAVERAM
        ld              c, 16

        call            MemFill         ; clear _AUD3WAVERAM ($FF30-$FF3F)

        ; initialize Low RAM variables

        ld              hl, ayAudTerm

        ld              [hl+], a        ; set ayAudTerm = 0
        ld              [hl], l         ; set ayMask    = $FF

        ld              hl, ayTonAFreq

        ld              a, $4F

        ld              [hl+], a
        ld              [hl], $03       ; set ayTonAFreq = $034F
        inc             l

        ld              [hl+], a
        ld              [hl], $03       ; set ayTonBFreq = $034F
        inc             l

        ld              [hl+], a
        ld              [hl], $03 << 1  ; set ayTonCFreq = $034F
        inc             l

        IF              CLICK_REDUCTION == 0
        xor             a
        ENDC
        ld              [hl+], a        ; set ayChnCVol <> 0

        IF              CLICK_REDUCTION == 1
        ld              a, $3F          ; disable TonABC & NoiABC
        ELSE
        xor             a               ; don't disable anything
        ENDC

        ld              [hl+], a        ; set ayMuted    = $3F / $00

        ld              a, $C0          ; disable EnvABC
        ld              [hl+], a
        ld              [hl], ayEnabledToAudTerm >> 8 + 1  ; setup ayEnabled

        ; initialize High RAM variables

        ld              hl, ayEnvPerNew

        xor             a
        ld              [hl+], a
        ld              [hl+], a        ; set ayEnvPerNew = 0
        ld              [hl], h         ; set ayGetEnv    = $FF (bit #4 set)
        inc             l
        ld              [hl], a         ; set ayEnvType   = 0

        ret

; **
; AY8914_Update:  Get new envelope variables (if ayGetEnv.4 = 1), update
;                 tone frequencies, update Wave RAM if necessary,
;                 enable/disable output channels, and calculate the
;                 current noise output level.
;
; Tables Used:    ayEnvTypeToHandler
;                 ayEnabledToAudTerm
;

AY8914_Update:

        ; -------------------------------------------------------------------
        ;   Get the latest envelope variables from ayEnvPerNew & ayEnvType
        ; -------------------------------------------------------------------

        ld              sp, ayEnvPerNew
        pop             de              ; de <- ayEnvPerNew
        pop             af              ; f  <- ayGetEnv (CY = 1 if TRUE)
                                        ; a  <- ayEnvType

        ld              sp, ayEnvPerCur + 2
        push            de              ; ayEnvPerCur <- ayEnvPerNew

        jr              nc, .UpdSndFreq ; if CY = 0, don't get env. variables

        ld              l, a
        ld              h, ayEnvTypeToHandler >> 8 ; hl -> ayEnvTypeToHandler
                                                   ;       [ ayEnvType ]

        ld              bc, $FF0F       ; ayEnvInc = -1, ayEnvVol = 15

        and             $08             ; is ATAK set (bit #3) ? (clear CY)
        jr              z, .NoATAK      ; no, so leave decay settings intact

        ld              bc, $0100       ; ayEnvInc = 1, ayEnvVol = 0
.NoATAK
        ldh             [ayGetEnv], a   ; update ayGetEnv (reset bit #4)

        push            de              ; update ayEnvCtr (=ayEnvPerCur)
        push            bc              ; update ayEnvInc and ayEnvVol

        ld              a, [hl+]
        ldh             [ayEnvEndHandler], a
        ld              a, [hl]
        ldh             [ayEnvEndHandler + 1], a  ; update ayEnvEndHandler

        ld              a, TACF_START | TACF_16KHZ
        ldh             [rTAC], a       ; re-start timer to update ayEnvVol

.UpdSndFreq

        ; -------------------------------------------------------------------
        ;    Update the frequency registers for Sound #1-#3 from Tone A-C
        ; -------------------------------------------------------------------

        ld              hl, ayTonAFreq

        ; update Sound #1's frequency from Tone A's frequency

        ld              a, [hl+]        ; a <- LOW byte of TonA's frequency
        ldh             [rAUD1LOW], a
        ld              a, [hl]         ; a <- HIGH byte of TonA's frequency
        ldh             [rAUD1HIGH], a

        and             $7F             ; clear Initial bit (in case set)
        ld              [hl+], a        ; update HIGH of TonA's frequency

        ; update Sound #2's frequency from Tone B's frequency

        ld              a, [hl+]        ; a <- LOW byte of TonB's frequency
        ldh             [rAUD2LOW], a
        ld              a, [hl]         ; a <- HIGH byte of TonB's frequency
        ldh             [rAUD2HIGH], a

        and             $7F             ; clear Initial bit (in case set)
        ld              [hl+], a        ; update HIGH of TonB's frequency

        ; update Sound #3's frequency from Tone C's frequency

        ld              a, [hl+]        ; a <- LOW byte of TonC's frequency
        ldh             [rAUD3LOW], a
        ld              a, [hl]         ; a <- HIGH byte of TonC's frequency

        rra                             ; CY <- Initial bit (volume changed)
        jr              nc, .UpdAud3Hi  ; vol. same, so just update rAUD3HIGH

        ; Sound #3's volume has changed, so modify the Wave Pattern RAM

        ld              d, a            ; save HIGH of TonC's frequency in d

        add             a, a            ; clear Initial bit for next time
        ld              [hl+], a        ; update HIGH of TonC's frequency

        ; set-up regs so Sound #3 is disabled for the shortest time possible

        ld              e, [hl]         ; e <- ChnC's vol. in LO & HI nybbles

        ld              bc, $80 << 8 + _AUD3WAVERAM & $FF + 12

        ld              hl, _AUD3WAVERAM + 4  ; hl -> samples 8-15 of WaveRAM

        ldh             [rAUD3ENA], a   ; stop Sound #3's output while
                                        ; updating the Wave Pattern RAM

        ld              a, e            ; a = ChnC's vol. in LO & HI nybbles

        ld              [hl+], a
        ld              [hl+], a
        ld              [hl+], a
        ld              [hl], a         ; store 8 samples of new volume

        ld              l, c            ; hl -> samples 24-31 of WaveRAM

        ld              [hl+], a
        ld              [hl+], a
        ld              [hl+], a
        ld              [hl], a         ; store 8 samples of new volume

        ld              a, b            ; set bit #7 (sound output OK)
        ldh             [rAUD3ENA], a   ; re-enable Sound #3's output
        or              d               ; a = HIGH of TonC's freq. + Initial
.UpdAud3Hi
        ldh             [rAUD3HIGH], a  ; update HIGH byte of frequency reg.

        ; -------------------------------------------------------------------
        ;     Set rAUDTERM based on the mixer and envelope enable bits
        ; -------------------------------------------------------------------

        ld              sp, ayMuted - 1
        pop             af              ; af <- ayMuted (f = ????)
        pop             de              ; de <- ayEnabled

        or              e
        ld              e, a            ; de -> ayEnabledToAudTerm

        ld              a, [de]         ; a <- new value for rAUDTERM

        ld              hl, ayAudTerm   ; hl -> ayAudTerm, followed by ayMask

        ld              [hl+], a        ; save value with SO2 bits intact
        and             [hl]            ; clear SO2 output enable bits if
                                        ; ayVolToAudVol[ ayEnvVol ] = $0F

        ldh             [rAUDTERM], a   ; update rAUDTERM

        ; -------------------------------------------------------------------
        ;     Update the volume level and polynomial counter for Sound #4
        ; -------------------------------------------------------------------

        xor             a               ; set maximum value = 0

        ; if NoiA is enabled, then check the current volume for ChnA

        bit             3, e            ; is NoiA enabled ?
        jr              nz, .ChkNoiB    ; no, so check if NoiB is enabled

        ldh             a, [rAUD1ENV]   ; a <- volume for ChnA (in bits 4-7)
.ChkNoiB
        ld              b, a            ; set the new maximum to [rAUD1ENV]

        ; if NoiB is enabled, then check the current volume for ChnB

        bit             4, e            ; is NoiB enabled ?
        jr              nz, .ChkNoiC    ; no, so check if NoiC is enabled

        ldh             a, [rAUD2ENV]   ; a <- volume for ChnB (in bits 4-7)
        cp              b               ; is it greater than the maximum ?
        jr              c, .ChkNoiC     ; no, so maximum doesn't change

        ld              b, a            ; set the new maximum to [rAUD2ENV]

.ChkNoiC

        ; if NoiC is enabled, then check the current volume for ChnC

        bit             5, e            ; is NoiC enabled ?
        jr              nz, .ChkNoiVol  ; no, so see if noise's vol. changed

        ld              a, [ayChnCVol]  ; a <- ChnC's vol (in HI & LO nybbles)
        cp              b               ; is it greater than the maximum ?
        jr              c, .ChkNoiVol   ; no, so maximum doesn't change

        and             $F0             ; clear bits 0-2 (would interfere)
        ld              b, a            ; set new maximum to [ayChnCVol]

.ChkNoiVol

        ld              a, b            ; a = new volume level for Sound #4
        or              a               ; is new volume = 0 ?
        jr              z, .UpdNoiPoly  ; then already muted, skip code below

        ld              hl, rAUD4ENV

        cp              [hl]            ; same as Snd #4's current volume ?
        jr              z, .UpdNoiPoly  ; yes, so don't bother to update it

        ld              [hl+], a        ; update rAUD4ENV w/ new volume

        ldh             a, [ayAudPoly]  ; a <- poly. based on noise period

        ld              [hl+], a        ; update rAUD4POLY
        ld              [hl], $80       ; update rAUD4GO (change volume)

        jp              AY8914_Ret      ; return to "caller"

.UpdNoiPoly

        ; update Sound #3's polynomial counter (based on noise frequency)

        ldh             a, [ayAudPoly]  ; a <- new value for polynomial
        ldh             [rAUD4POLY], a  ; update rAUD4POLY

        jp              AY8914_Ret      ; return to "caller"
