; **
; Intellivision for GameBoy Color v1.1
; Copyright(C) 2000,03 By Carl Mueller, Jr.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

; ---------------------------------------------------------------------------
;                   .::** STIC Emulation Routines **::.
; ---------------------------------------------------------------------------

ST_X7:  ld              hl, MO + MO_RECSIZE * 7
        jr              STIC_XPosition

ST_X6:  ld              hl, MO + MO_RECSIZE * 6
        jr              STIC_XPosition

ST_X5:  ld              hl, MO + MO_RECSIZE * 5
        jr              STIC_XPosition

ST_X4:  ld              hl, MO + MO_RECSIZE * 4
        jr              STIC_XPosition

ST_X3:  ld              hl, MO + MO_RECSIZE * 3
        jr              STIC_XPosition

ST_X2:  ld              hl, MO + MO_RECSIZE * 2
        jr              STIC_XPosition

ST_X1:  ld              hl, MO + MO_RECSIZE * 1
        jr              STIC_XPosition

ST_X0:  ld              hl, MO + MO_RECSIZE * 0

STIC_XPosition:                          ; STIC Register #0-#7 @ $0000-$0007

        ; $00 + n : X position register, moving object #n.
        ;
        ;   15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
        ;    ?  ?  ?  ?  ? xs  v  c  x  x  x  x  x  x  x  x

        ld              a, [hl]         ; a <- MO attribute byte #0

        xor             d
        and             $FC             ; clear V and C
        xor             d               ; set V and C
        or              $80             ; set ReAnalyze

        ld              [hl+], a        ; store out new MO attribute byte #0

        ld              [hl], e         ; store out new X Coord.
        inc             l
        inc             l               ; skip Y-Coord.

        ld              a, [hl]         ; a <- MO attribute byte #1

        xor             d               ; changed bits will be set
        and             $04             ; did XSIZE change ?
        jr              z, .Exit        ; no, so just return

        xor             [hl]            ; flip XSIZE (set it to new value)
        or              $80             ; set ReCopy

        ld              [hl], a         ; store new MO attribute byte #1
.Exit
        reti                            ; return, enabling interrupts

ST_A7:  ld              hl, MO + MO_RECSIZE * 7
        jr              STIC_Attribute

ST_A6:  ld              hl, MO + MO_RECSIZE * 6
        jr              STIC_Attribute

ST_A5:  ld              hl, MO + MO_RECSIZE * 5
        jr              STIC_Attribute

ST_A4:  ld              hl, MO + MO_RECSIZE * 4
        jr              STIC_Attribute

ST_A3:  ld              hl, MO + MO_RECSIZE * 3
        jr              STIC_Attribute

ST_A2:  ld              hl, MO + MO_RECSIZE * 2
        jr              STIC_Attribute

ST_A1:  ld              hl, MO + MO_RECSIZE * 1
        jr              STIC_Attribute

ST_A0:  ld              hl, MO + MO_RECSIZE * 0

STIC_Attribute:                         ; STIC Register #16-#23 @ $0010-$0017

        ; $10 + n : Attribute register, moving object #n.
        ;
        ;   15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
        ;    ?  ?  p  F  t  n  n  n  n  n  n  n  n  f  f  f

        ld              a, d            ; a = HIGH byte of Attribute reg.
        scf                             ; * make sure CY = 1
        rra                             ; put P in bit #4, bit #7 = CY

        xor             [hl]            ; xor it w/ old MO attribute byte #0
        and             $90             ; isolate P and ReAnalyze
        xor             [hl]            ; set P and ReAnalyze

        ld              [hl+], a        ; store out new MO attribute byte #0

        inc             l               ; skip X-Coord.
        inc             l               ; skip Y-Coord.
        inc             l               ; skip MO attribute byte #1

        ; compute the card's new data address & compare it w/ the old

        ld              a, e            ; a = LOW byte of Attribute reg.
        and             $F8             ; isolate LSBs of card index

        cp              [hl]            ; is it = LOW of old data address ?
        jr              z, .HiCheck     ; yes, so check HIGH too to make sure

        ; since the LOW byte of the data address differs, the MO must now
        ; be using a different (set of) card(s), so store out the new data
        ; address and set the MO's ReCopy flag (so it's queued in MOVELIST)

        ld              [hl+], a        ; store out LOW of new data address

        ld              a, d
        and             $07             ; isolate MSBs of card index
        add             a, GROM_8 >> 8  ; assume it will be in GROM (0-255)

        bit             3, d            ; is GRAM toggle set ?
        jr              z, .InGROM_1    ; no, so leave it in GROM

        and             $01             ; only 1 MSB if card comes from GRAM
        add             a, GRAM_8 >> 8  ; card is in GRAM (0-63)
.InGROM_1
        ld              [hl-], a        ; store out HIGH of new data address

        dec             l               ; pos @ MO attribute byte #1

        set             7, [hl]         ; set ReCopy flag

        ld              a, l            ; a = ..mmm.11 (MO # * 8 + 3)
        srl             a               ;   = ...mmm.1 and CY = 1 *
        rra                             ;   = 1...mmm. and CY = 1
        rra                             ;   = 11...mmm

        jr              .SetColor       ; *

.HiCheck

        ; LOW byte of the data address was the same, but the HIGH byte
        ; may still differ, so check it to make sure

        inc             l               ; pos @ HIGH byte of data address

        ld              a, d
        and             $07             ; isolate MSBs of card index
        add             a, GROM_8 >> 8  ; assume it will be in GROM (0-255)

        bit             3, d            ; is GRAM toggle set ?
        jr              z, .InGROM_2

        and             $01             ; only 1 MSB if card comes from GRAM
        add             a, GRAM_8 >> 8  ; card is in GRAM (0-63)
.InGROM_2
        cp              [hl]            ; is it = HIGH of old data address ?
        jr              nz, .InGROM_1

        ld              a, l            ; a = ..mmm1.1 (MO # * 8 + 5)
        rra                             ;   = ...mmm1. and CY = 1
        rra                             ;   = 1...mmm1
        rrca                            ;   = 11...mmm

.SetColor

        ld              l, a            ; hl -> MO_FgCols[ MO # ]

        ld              a, d            ; a = HIGH byte of Attribute reg.
        and             $10             ; isolate F (MSB of color) in bit #4

        rrca                            ; put F in bit #3
        xor             e
        and             $F8             ; clear fff (LSBs of color)
        xor             e               ; set fff

        swap            a               ; put Ffff in HIGH nybble
        inc             a               ; set ReCopy bit

        xor             [hl]
        and             $FB
        xor             [hl]            ; * preserve bit #2 (YS << 1 + YF <> 0)

        ld              [hl], a         ; update FgCol byte

        reti                            ; return, enabling interrupts

STIC_DisplayEnable:                     ; STIC Register #32 @ $0020

        ; the CP1610 program must write to $20 in VBLANK to enable display

        ld              a, 1
        ld              [STIC_Display], a

        reti                            ; return, enabling interrupts

STIC_GraphicsMode:                      ; STIC Register #33 @ $0021

        ; writes to $21 set the STIC in FG/BG mode -- reads set it in CS mode

        ld              a, 1            ; a = 1 (FG/BG mode)
        ld              [STIC_Mode_New], a

        ld              hl, BW_BW + 1   ; AP_BW + 1

        ld              a, BktWrite_FGBG & $FF
        ld              [hl+], a
        ld              [hl], BktWrite_FGBG >> 8

        reti                            ; return, enabling interrupts

ST_C3:  ld              hl, ColorStack + 3
        jr              STIC_ColorStack

ST_C2:  ld              hl, ColorStack + 2
        jr              STIC_ColorStack

ST_C1:  ld              hl, ColorStack + 1
        jr              STIC_ColorStack

ST_C0:  ld              hl, ColorStack + 0

STIC_ColorStack:                        ; STIC Register #40-#43 @ $0028-$002B

        ld              [SP_1], sp
        ld              sp, SP_1        ; use STACK_1 for below

        ei                              ; interrupts can't be off for so long

        ; write ColNum << 1 to "ColorStack", used by CS_BktWrite, etc.

        ld              a, e            ; a = ????cccc
        and             $0F             ;   = ....cccc
        add             a, a            ;   = ...cccc. (ColNum << 1)

        ld              [hl], a         ; ColorStack[ n ] <- ColNum << 1

        ; set new BG colors for MOs that use this ColorStack element

        rrca                            ;   = ....cccc (ColNum)
        add             OBJPAL >> 7
        ld              e, a            ; e = OBJPAL >> 7 + ColNum

        ld              a, l            ; a = LOW byte of ColorStack[ n ]

        ld              hl, MO_CsCols   ; hl -> MO_CsCols[ 0 ]

        ld              d, 8            ; check the bkg. colors of 8 MOs
.CsLoop
        cp              [hl]            ; does MO use this CS index # ?
        jr              nz, .CsCont     ; no, so don't change its bkg. color

        res             4, l            ; hl -> MO_FgCols[ MO # ]

        set             0, [hl]         ; set ReCopy

        set             3, l            ; hl -> MO_BgCols[ MO # ]

        ld              [hl], e         ; store out new BgCol byte

        res             3, l
        set             4, l            ; hl -> MO_CsCols[ MO # ]
.CsCont
        inc             l               ; pos @ next MO's bkg. color

        dec             d               ; have we checked all 8 MOs ?
        jr              nz, .CsLoop     ; no, so keep looping

        ; ^ 9 + 6 + 9 + 79-175 = 103-199 cycles

        ld              a, 3            ; a = #/WRAM bank where CARD is
        ldh             [cSVBK], a      ; note new WRAM bank for ISRs
        ldh             [rSVBK], a      ; switch to new WRAM bank

        ld              hl, CARD + 2
        ld              e, CARD_RECSIZE
        ld              a, 240          ; do 240 cards
.CardLoop
        set             7, [hl]         ; set Re-analyze (bit #7)

        add             hl, de          ; advance to next CARD record

        dec             a
        jr              nz, .CardLoop

        ld              [hl], e         ; set BktAnalyze_Flag <> 0

        inc             a               ; * a = #/WRAM bank w/ STIC, etc.
        ldh             [cSVBK], a      ; note new WRAM bank for ISRs
        ldh             [rSVBK], a      ; switch to new WRAM bank

        ; ^ 15 + 2159 + 9 = 2183 cycles

        di                              ; disable interrupts to restore stack

        pop             hl
        ld              sp, hl

        reti                            ; return, enabling interrupts

STIC_BorderColor:                       ; STIC Register #44 @ $002C

        ; look up new border color's RGB value and store it in [BrdrColRGB]

        ld              a, e              ; a = ????cccc
        and             $0F               ;   = ....cccc
        add             a, a              ;   = ...cccc. (ColNum << 1)
        add             ColorToRGB & $FF

        ld              l, a
        ld              h, ColorToRGB >> 8  ; hl -> ColorToRGB[ ColNum << 1 ]

        ld              a, [hl+]        ; a <- LOW byte of RGB color value
        ld              d, [hl]         ; d <- HIGH byte of RGB color value

        ld              l, BrdrColRGB & $FF

        ld              [hl+], a        ; store out LOW of RGB color value
        ld              [hl], d         ; store out HIGH of RGB color value

;        ld              hl, WINPAL_0
;
;        ld              [hl+], a        ; store out LOW of RGB color value
;        ld              [hl], d         ; store out HIGH of RGB color value

        ; ^ 26 cycles, 20 bytes

        reti                            ; return, enabling interrupts

STIC_HorizontalDelay:                   ; STIC Register #48 @ $0030

        ld              a, e            ; a = ?????ddd
        and             $07             ;   = .....ddd (H-Delay)
        IF              BkgOrgX != 0
        add             BkgOrgX
        ENDC

        cpl
        inc             a               ; a = -1 * ( BkgOrgX + H-Delay )

        ld              [rSCX_New], a   ; store it out to Scroll X "new"

        jr              STIC_SetReAnalyze

STIC_VerticalDelay:                     ; STIC Register #49 @ $0031

        ld              a, e            ; a = ?????ddd
        and             $07             ;   = .....ddd (V-Delay)
        IF              BkgOrgY != 0
        add             BkgOrgY
        ENDC

        cpl
        inc             a               ; a = -1 * ( BkgOrgY + V-Delay )

        ld              [rSCY_New], a   ; store it out to Scroll Y "new"

STIC_SetReAnalyze:

        ; since MOs move w/ the background, their positions need updating

        ld              hl, MO
        ld              de, MO_RECSIZE

        REPT            7               ; do 1st 7 moving objects

        set             7, [hl]         ; set ReAnalyze
        add             hl, de          ; advance to next MO record

        ENDR

        set             7, [hl]         ; set ReAnalyze

        ; ^ 15/12 + 6 + 38 = 59/56 cycles, 52 bytes

        reti                            ; return, enabling interrupts

STIC_BlockOut:                          ; STIC Register #50 @ $0032

        ; update LY_WinOff & BC_Active according to new value of H-BlockOut

        ld              a, e            ; a = ??????hv
        and             $02             ;   = ......h.
        add             a, a            ;   = .....h..
        add             a, a            ;   = ....h... (H-BlockOut << 3)

        add             BkgOrgY - 1     ; a = BkgOrgY - 1 or BkgOrgY - 1 + 8

        ld              [LY_WinOff], a  ; LY_WinOff <- 1 row b4 window = OFF

        ld              a, e            ; a = ??????hv
        and             $01             ;   = .......v
        swap            a               ;   = ...v.... (V-BlockOut << 4)

        ld              d, a            ; save V-Blockout << 4 in d

        ldh             [BC_Active], a  ; let VBlankFunc know column's status

        ; re-position column in LocalOAM according to new value of V-BlockOut

        ld              a, 5
        ldh             [rSVBK], a      ; switch to WRAM bank #5 (LocalOAM)

        ld              a, BkgOrgY

SPR_NUM SET             0
        REPT            6               ; column composed of 6 sprites

        add             d
        ld              [LocalOAM + SPR_NUM * 4], a  ; set Y-Coordinate

SPR_NUM SET             SPR_NUM + 1
        ENDR

        ld              a, 1
        ldh             [rSVBK], a      ; switch back to WRAM bank #1

        ; ^ 20 + 7 + 30 + 5 = 62 cycles, 52 bytes

ST_RI:  reti                            ; return, enabling interrupts

